import time

import pytest
from flask import current_app
from flask import json

from authlib.common.urls import url_decode
from authlib.common.urls import url_encode
from authlib.common.urls import urlparse
from authlib.jose import JsonWebToken
from authlib.jose import jwt
from authlib.oauth2.rfc6749.grants import (
    AuthorizationCodeGrant as _AuthorizationCodeGrant,
)
from authlib.oidc.core import CodeIDToken
from authlib.oidc.core.grants import OpenIDCode as _OpenIDCode
from tests.util import read_file_path

from .models import Client
from .models import CodeGrantMixin
from .models import exists_nonce
from .models import save_authorization_code
from .oauth2_server import create_basic_header


@pytest.fixture(autouse=True)
def client(client, db):
    client.set_client_metadata(
        {
            "redirect_uris": ["https://client.test"],
            "scope": "openid profile address",
            "response_types": ["code"],
            "grant_types": ["authorization_code"],
        }
    )
    db.session.add(client)
    db.session.commit()
    return client


@pytest.fixture(autouse=True)
def server(server, app):
    app.config.update(
        {
            "OAUTH2_JWT_ISS": "Authlib",
            "OAUTH2_JWT_KEY": "secret",
            "OAUTH2_JWT_ALG": "HS256",
        }
    )
    return server


def register_oidc_code_grant(server, require_nonce=False):
    class AuthorizationCodeGrant(CodeGrantMixin, _AuthorizationCodeGrant):
        def save_authorization_code(self, code, request):
            return save_authorization_code(code, request)

    class OpenIDCode(_OpenIDCode):
        def get_jwt_config(self, grant, client):
            key = current_app.config.get("OAUTH2_JWT_KEY")
            alg = current_app.config.get("OAUTH2_JWT_ALG")
            iss = current_app.config.get("OAUTH2_JWT_ISS")
            return dict(key=key, alg=alg, iss=iss, exp=3600)

        def exists_nonce(self, nonce, request):
            return exists_nonce(nonce, request)

        def generate_user_info(self, user, scopes):
            return user.generate_user_info(scopes)

    server.register_grant(
        AuthorizationCodeGrant, [OpenIDCode(require_nonce=require_nonce)]
    )


def test_authorize_token(test_client, server):
    register_oidc_code_grant(
        server,
    )
    auth_request_time = time.time()
    rv = test_client.post(
        "/oauth/authorize",
        data={
            "response_type": "code",
            "client_id": "client-id",
            "state": "bar",
            "scope": "openid profile",
            "redirect_uri": "https://client.test",
            "user_id": "1",
        },
    )
    assert "code=" in rv.location

    params = dict(url_decode(urlparse.urlparse(rv.location).query))
    assert params["state"] == "bar"

    code = params["code"]
    headers = create_basic_header("client-id", "client-secret")
    rv = test_client.post(
        "/oauth/token",
        data={
            "grant_type": "authorization_code",
            "redirect_uri": "https://client.test",
            "code": code,
        },
        headers=headers,
    )
    resp = json.loads(rv.data)
    assert "access_token" in resp
    assert "id_token" in resp

    claims = jwt.decode(
        resp["id_token"],
        "secret",
        claims_cls=CodeIDToken,
        claims_options={"iss": {"value": "Authlib"}},
    )
    claims.validate()
    assert claims["auth_time"] >= int(auth_request_time)
    assert claims["acr"] == "urn:mace:incommon:iap:silver"
    assert claims["amr"] == ["pwd", "otp"]


def test_pure_code_flow(test_client, server):
    register_oidc_code_grant(
        server,
    )
    rv = test_client.post(
        "/oauth/authorize",
        data={
            "response_type": "code",
            "client_id": "client-id",
            "state": "bar",
            "scope": "profile",
            "redirect_uri": "https://client.test",
            "user_id": "1",
        },
    )
    assert "code=" in rv.location

    params = dict(url_decode(urlparse.urlparse(rv.location).query))
    assert params["state"] == "bar"

    code = params["code"]
    headers = create_basic_header("client-id", "client-secret")
    rv = test_client.post(
        "/oauth/token",
        data={
            "grant_type": "authorization_code",
            "redirect_uri": "https://client.test",
            "code": code,
        },
        headers=headers,
    )
    resp = json.loads(rv.data)
    assert "access_token" in resp
    assert "id_token" not in resp


def test_require_nonce(test_client, server):
    register_oidc_code_grant(server, require_nonce=True)
    rv = test_client.post(
        "/oauth/authorize",
        data={
            "response_type": "code",
            "client_id": "client-id",
            "user_id": "1",
            "state": "bar",
            "scope": "openid profile",
            "redirect_uri": "https://client.test",
        },
    )
    params = dict(url_decode(urlparse.urlparse(rv.location).query))
    assert params["error"] == "invalid_request"
    assert params["error_description"] == "Missing 'nonce' in request."


def test_nonce_replay(test_client, server):
    register_oidc_code_grant(
        server,
    )
    data = {
        "response_type": "code",
        "client_id": "client-id",
        "user_id": "1",
        "state": "bar",
        "nonce": "abc",
        "scope": "openid profile",
        "redirect_uri": "https://client.test",
    }
    rv = test_client.post("/oauth/authorize", data=data)
    assert "code=" in rv.location

    rv = test_client.post("/oauth/authorize", data=data)
    assert "error=" in rv.location


def test_prompt(test_client, server):
    register_oidc_code_grant(
        server,
    )
    params = [
        ("response_type", "code"),
        ("client_id", "client-id"),
        ("state", "bar"),
        ("nonce", "abc"),
        ("scope", "openid profile"),
        ("redirect_uri", "https://client.test"),
    ]
    query = url_encode(params)
    rv = test_client.get("/oauth/authorize?" + query)
    assert rv.data == b"login"

    query = url_encode(params + [("user_id", "1")])
    rv = test_client.get("/oauth/authorize?" + query)
    assert rv.data == b"ok"

    query = url_encode(params + [("prompt", "login")])
    rv = test_client.get("/oauth/authorize?" + query)
    assert rv.data == b"login"

    query = url_encode(params + [("user_id", "1"), ("prompt", "login")])
    rv = test_client.get("/oauth/authorize?" + query)
    assert rv.data == b"login"


def test_prompt_none_not_logged(test_client, server):
    register_oidc_code_grant(
        server,
    )
    params = [
        ("response_type", "code"),
        ("client_id", "client-id"),
        ("state", "bar"),
        ("nonce", "abc"),
        ("scope", "openid profile"),
        ("redirect_uri", "https://client.test"),
        ("prompt", "none"),
    ]
    query = url_encode(params)
    rv = test_client.get("/oauth/authorize?" + query)

    params = dict(url_decode(urlparse.urlparse(rv.location).query))
    assert params["error"] == "login_required"
    assert params["state"] == "bar"


def test_client_metadata_custom_alg(test_client, server, client, db, app):
    """If the client metadata 'id_token_signed_response_alg' is defined,
    it should be used to sign id_tokens."""
    register_oidc_code_grant(
        server,
    )
    client.set_client_metadata(
        {
            "redirect_uris": ["https://client.test"],
            "scope": "openid profile address",
            "response_types": ["code"],
            "grant_types": ["authorization_code"],
            "id_token_signed_response_alg": "HS384",
        }
    )
    db.session.add(client)
    db.session.commit()
    del app.config["OAUTH2_JWT_ALG"]

    rv = test_client.post(
        "/oauth/authorize",
        data={
            "response_type": "code",
            "client_id": "client-id",
            "state": "bar",
            "scope": "openid profile",
            "redirect_uri": "https://client.test",
            "user_id": "1",
        },
    )
    params = dict(url_decode(urlparse.urlparse(rv.location).query))
    code = params["code"]
    headers = create_basic_header("client-id", "client-secret")
    rv = test_client.post(
        "/oauth/token",
        data={
            "grant_type": "authorization_code",
            "redirect_uri": "https://client.test",
            "code": code,
        },
        headers=headers,
    )
    resp = json.loads(rv.data)
    claims = jwt.decode(
        resp["id_token"],
        "secret",
        claims_cls=CodeIDToken,
        claims_options={"iss": {"value": "Authlib"}},
    )
    claims.validate()
    assert claims.header["alg"] == "HS384"


def test_client_metadata_alg_none(test_client, server, app, db, client):
    """The 'none' 'id_token_signed_response_alg' alg should be
    supported in non implicit flows."""
    register_oidc_code_grant(
        server,
    )
    client.set_client_metadata(
        {
            "redirect_uris": ["https://client.test"],
            "scope": "openid profile address",
            "response_types": ["code"],
            "grant_types": ["authorization_code"],
            "id_token_signed_response_alg": "none",
        }
    )
    db.session.add(client)
    db.session.commit()

    del app.config["OAUTH2_JWT_ALG"]
    rv = test_client.post(
        "/oauth/authorize",
        data={
            "response_type": "code",
            "client_id": "client-id",
            "state": "bar",
            "scope": "openid profile",
            "redirect_uri": "https://client.test",
            "user_id": "1",
        },
    )
    params = dict(url_decode(urlparse.urlparse(rv.location).query))
    code = params["code"]
    headers = create_basic_header("client-id", "client-secret")
    rv = test_client.post(
        "/oauth/token",
        data={
            "grant_type": "authorization_code",
            "redirect_uri": "https://client.test",
            "code": code,
        },
        headers=headers,
    )
    resp = json.loads(rv.data)
    jwt_none = JsonWebToken(["none"])
    claims = jwt_none.decode(
        resp["id_token"],
        "secret",
        claims_cls=CodeIDToken,
        claims_options={"iss": {"value": "Authlib"}},
    )
    claims.validate()
    assert claims.header["alg"] == "none"


@pytest.mark.parametrize(
    "alg, private_key, public_key",
    [
        (
            "RS256",
            read_file_path("jwk_private.json"),
            read_file_path("jwk_public.json"),
        ),
        (
            "PS256",
            read_file_path("jwks_private.json"),
            read_file_path("jwks_public.json"),
        ),
        (
            "ES512",
            read_file_path("secp521r1-private.json"),
            read_file_path("secp521r1-public.json"),
        ),
        (
            "RS256",
            read_file_path("rsa_private.pem"),
            read_file_path("rsa_public.pem"),
        ),
    ],
)
def test_authorize_token_algs(test_client, server, app, alg, private_key, public_key):
    # generate refresh token
    app.config["OAUTH2_JWT_KEY"] = private_key
    app.config["OAUTH2_JWT_ALG"] = alg
    register_oidc_code_grant(
        server,
    )
    rv = test_client.post(
        "/oauth/authorize",
        data={
            "response_type": "code",
            "client_id": "client-id",
            "state": "bar",
            "scope": "openid profile",
            "redirect_uri": "https://client.test",
            "user_id": "1",
        },
    )
    assert "code=" in rv.location

    params = dict(url_decode(urlparse.urlparse(rv.location).query))
    assert params["state"] == "bar"

    code = params["code"]
    headers = create_basic_header("client-id", "client-secret")
    rv = test_client.post(
        "/oauth/token",
        data={
            "grant_type": "authorization_code",
            "redirect_uri": "https://client.test",
            "code": code,
        },
        headers=headers,
    )
    resp = json.loads(rv.data)
    assert "access_token" in resp
    assert "id_token" in resp

    claims = jwt.decode(
        resp["id_token"],
        public_key,
        claims_cls=CodeIDToken,
        claims_options={"iss": {"value": "Authlib"}},
    )
    claims.validate()


def test_deprecated_get_jwt_config_signature(test_client, server, db, user):
    """Using the old get_jwt_config(self, grant) signature should emit a DeprecationWarning."""

    class DeprecatedOpenIDCode(_OpenIDCode):
        def get_jwt_config(self, grant):
            return {"key": "secret", "alg": "HS256", "iss": "Authlib", "exp": 3600}

        def exists_nonce(self, nonce, request):
            return exists_nonce(nonce, request)

        def generate_user_info(self, user, scopes):
            return user.generate_user_info(scopes)

    class AuthorizationCodeGrant(CodeGrantMixin, _AuthorizationCodeGrant):
        def save_authorization_code(self, code, request):
            return save_authorization_code(code, request)

    server.register_grant(AuthorizationCodeGrant, [DeprecatedOpenIDCode()])

    client = Client(
        user_id=user.id,
        client_id="deprecated-client",
        client_secret="secret",
    )
    client.set_client_metadata(
        {
            "redirect_uris": ["https://client.test"],
            "scope": "openid profile",
            "response_types": ["code"],
            "grant_types": ["authorization_code"],
        }
    )
    db.session.add(client)
    db.session.commit()

    rv = test_client.post(
        "/oauth/authorize",
        data={
            "response_type": "code",
            "client_id": "deprecated-client",
            "state": "bar",
            "scope": "openid profile",
            "redirect_uri": "https://client.test",
            "user_id": "1",
        },
    )
    params = dict(url_decode(urlparse.urlparse(rv.location).query))
    code = params["code"]

    with pytest.warns(DeprecationWarning, match="get_jwt_config.*version 1.8"):
        test_client.post(
            "/oauth/token",
            data={
                "grant_type": "authorization_code",
                "redirect_uri": "https://client.test",
                "code": code,
            },
            headers=create_basic_header("deprecated-client", "secret"),
        )



from flask_sqlalchemy import SQLAlchemy

from authlib.integrations.sqla_oauth2 import OAuth2AuthorizationCodeMixin
from authlib.integrations.sqla_oauth2 import OAuth2ClientMixin
from authlib.integrations.sqla_oauth2 import OAuth2TokenMixin
from authlib.oidc.core import UserInfo

db = SQLAlchemy()


class User(db.Model):
    id = db.Column(db.Integer, primary_key=True)
    username = db.Column(db.String(40), unique=True, nullable=False)

    def get_user_id(self):
        return self.id

    def check_password(self, password):
        return password != "wrong"

    def generate_user_info(self, scopes=None):
        profile = {
            "sub": str(self.id),
            "name": self.username,
            "given_name": "Jane",
            "family_name": "Doe",
            "middle_name": "Middle",
            "nickname": "Jany",
            "preferred_username": "j.doe",
            "profile": "https://resource.test/janedoe",
            "picture": "https://resource.test/janedoe/me.jpg",
            "website": "https://resource.test",
            "email": "janedoe@example.com",
            "email_verified": True,
            "gender": "female",
            "birthdate": "2000-12-01",
            "zoneinfo": "Europe/Paris",
            "locale": "fr-FR",
            "phone_number": "+1 (425) 555-1212",
            "phone_number_verified": False,
            "address": {
                "formatted": "742 Evergreen Terrace, Springfield",
                "street_address": "742 Evergreen Terrace",
                "locality": "Springfield",
                "region": "Unknown",
                "postal_code": "1245",
                "country": "USA",
            },
            "updated_at": 1745315119,
        }
        return UserInfo(profile)


class Client(db.Model, OAuth2ClientMixin):
    id = db.Column(db.Integer, primary_key=True)
    user_id = db.Column(db.Integer, db.ForeignKey("user.id", ondelete="CASCADE"))
    user = db.relationship("User")


class AuthorizationCode(db.Model, OAuth2AuthorizationCodeMixin):
    id = db.Column(db.Integer, primary_key=True)
    user_id = db.Column(db.Integer, nullable=False)

    @property
    def user(self):
        return db.session.get(User, self.user_id)


class Token(db.Model, OAuth2TokenMixin):
    id = db.Column(db.Integer, primary_key=True)
    user_id = db.Column(db.Integer, db.ForeignKey("user.id", ondelete="CASCADE"))
    user = db.relationship("User")

    def is_refresh_token_active(self):
        return not self.refresh_token_revoked_at

    def get_client(self):
        return db.session.query(Client).filter_by(client_id=self.client_id).one()

    def get_user(self):
        return self.user


class CodeGrantMixin:
    def query_authorization_code(self, code, client):
        item = AuthorizationCode.query.filter_by(
            code=code, client_id=client.client_id
        ).first()
        if item and not item.is_expired():
            return item

    def delete_authorization_code(self, authorization_code):
        db.session.delete(authorization_code)
        db.session.commit()

    def authenticate_user(self, authorization_code):
        return db.session.get(User, authorization_code.user_id)


def save_authorization_code(code, request):
    client = request.client
    auth_code = AuthorizationCode(
        code=code,
        client_id=client.client_id,
        redirect_uri=request.payload.redirect_uri,
        scope=request.payload.scope,
        nonce=request.payload.data.get("nonce"),
        user_id=request.user.id,
        code_challenge=request.payload.data.get("code_challenge"),
        code_challenge_method=request.payload.data.get("code_challenge_method"),
        acr="urn:mace:incommon:iap:silver",
        amr="pwd otp",
    )
    db.session.add(auth_code)
    db.session.commit()
    return auth_code


def exists_nonce(nonce, request):
    exists = AuthorizationCode.query.filter_by(
        client_id=request.payload.client_id, nonce=nonce
    ).first()
    return bool(exists)



import pytest
from flask import json

from authlib.common.urls import add_params_to_uri
from authlib.oauth2.rfc6749.grants import (
    ResourceOwnerPasswordCredentialsGrant as _PasswordGrant,
)
from authlib.oidc.core import OpenIDToken

from .models import User
from .oauth2_server import create_basic_header


@pytest.fixture(autouse=True)
def client(client, db):
    client.set_client_metadata(
        {
            "scope": "openid profile",
            "grant_types": ["password"],
            "redirect_uris": ["https://client.test/authorized"],
        }
    )
    db.session.add(client)
    db.session.commit()
    return client


class IDToken(OpenIDToken):
    def get_jwt_config(self, grant, client):
        return {
            "iss": "Authlib",
            "key": "secret",
            "alg": "HS256",
        }

    def generate_user_info(self, user, scopes):
        return user.generate_user_info(scopes)


class PasswordGrant(_PasswordGrant):
    def authenticate_user(self, username, password):
        user = User.query.filter_by(username=username).first()
        if user.check_password(password):
            return user


def register_password_grant(server, extensions=None):
    server.register_grant(PasswordGrant, extensions)


def test_invalid_client(test_client, server):
    register_password_grant(
        server,
    )
    rv = test_client.post(
        "/oauth/token",
        data={
            "grant_type": "password",
            "username": "foo",
            "password": "ok",
        },
    )
    resp = json.loads(rv.data)
    assert resp["error"] == "invalid_client"

    headers = create_basic_header("client-id", "invalid-secret")
    rv = test_client.post(
        "/oauth/token",
        data={
            "grant_type": "password",
            "username": "foo",
            "password": "ok",
        },
        headers=headers,
    )
    resp = json.loads(rv.data)
    assert resp["error"] == "invalid_client"


def test_invalid_scope(test_client, server):
    register_password_grant(
        server,
    )
    server.scopes_supported = ["profile"]
    headers = create_basic_header("client-id", "client-secret")
    rv = test_client.post(
        "/oauth/token",
        data={
            "grant_type": "password",
            "username": "foo",
            "password": "ok",
            "scope": "invalid",
        },
        headers=headers,
    )
    resp = json.loads(rv.data)
    assert resp["error"] == "invalid_scope"


def test_invalid_request(test_client, server):
    register_password_grant(
        server,
    )
    headers = create_basic_header("client-id", "client-secret")

    rv = test_client.get(
        add_params_to_uri(
            "/oauth/token",
            {
                "grant_type": "password",
            },
        ),
        headers=headers,
    )
    resp = json.loads(rv.data)
    assert resp["error"] == "unsupported_grant_type"

    rv = test_client.post(
        "/oauth/token",
        data={
            "grant_type": "password",
        },
        headers=headers,
    )
    resp = json.loads(rv.data)
    assert resp["error"] == "invalid_request"

    rv = test_client.post(
        "/oauth/token",
        data={
            "grant_type": "password",
            "username": "foo",
        },
        headers=headers,
    )
    resp = json.loads(rv.data)
    assert resp["error"] == "invalid_request"

    rv = test_client.post(
        "/oauth/token",
        data={
            "grant_type": "password",
            "username": "foo",
            "password": "wrong",
        },
        headers=headers,
    )
    resp = json.loads(rv.data)
    assert resp["error"] == "invalid_request"


def test_invalid_grant_type(test_client, server, db, client):
    register_password_grant(server)
    client.set_client_metadata(
        {
            "scope": "openid profile",
            "grant_types": ["invalid"],
            "redirect_uris": ["https://client.test/authorized"],
        }
    )
    db.session.add(client)
    db.session.commit()

    headers = create_basic_header("client-id", "client-secret")
    rv = test_client.post(
        "/oauth/token",
        data={
            "grant_type": "password",
            "username": "foo",
            "password": "ok",
        },
        headers=headers,
    )
    resp = json.loads(rv.data)
    assert resp["error"] == "unauthorized_client"


def test_authorize_token(test_client, server):
    register_password_grant(
        server,
    )
    headers = create_basic_header("client-id", "client-secret")
    rv = test_client.post(
        "/oauth/token",
        data={
            "grant_type": "password",
            "username": "foo",
            "password": "ok",
        },
        headers=headers,
    )
    resp = json.loads(rv.data)
    assert "access_token" in resp


def test_token_generator(test_client, server, app):
    m = "tests.flask.test_oauth2.oauth2_server:token_generator"
    app.config.update({"OAUTH2_ACCESS_TOKEN_GENERATOR": m})
    server.load_config(app.config)
    register_password_grant(server)
    headers = create_basic_header("client-id", "client-secret")
    rv = test_client.post(
        "/oauth/token",
        data={
            "grant_type": "password",
            "username": "foo",
            "password": "ok",
        },
        headers=headers,
    )
    resp = json.loads(rv.data)
    assert "access_token" in resp
    assert "c-password.1." in resp["access_token"]


def test_custom_expires_in(test_client, server, app):
    app.config.update({"OAUTH2_TOKEN_EXPIRES_IN": {"password": 1800}})
    server.load_config(app.config)
    register_password_grant(server)
    headers = create_basic_header("client-id", "client-secret")
    rv = test_client.post(
        "/oauth/token",
        data={
            "grant_type": "password",
            "username": "foo",
            "password": "ok",
        },
        headers=headers,
    )
    resp = json.loads(rv.data)
    assert "access_token" in resp
    assert resp["expires_in"] == 1800


def test_id_token_extension(test_client, server):
    register_password_grant(server, extensions=[IDToken()])
    headers = create_basic_header("client-id", "client-secret")
    rv = test_client.post(
        "/oauth/token",
        data={
            "grant_type": "password",
            "username": "foo",
            "password": "ok",
            "scope": "openid profile",
        },
        headers=headers,
    )
    resp = json.loads(rv.data)
    assert "access_token" in resp
    assert "id_token" in resp



import time

import pytest
from flask import json
from flask import jsonify

from authlib.common.security import generate_token
from authlib.integrations.flask_oauth2 import ResourceProtector
from authlib.integrations.flask_oauth2 import current_token
from authlib.jose import jwt
from authlib.oauth2.rfc9068 import JWTBearerTokenValidator
from tests.util import read_file_path

from ..models import Token
from ..models import User
from ..models import db

issuer = "https://provider.test/"
resource_server = "resource-server-id"


@pytest.fixture(autouse=True)
def token_validator(jwks):
    class MyJWTBearerTokenValidator(JWTBearerTokenValidator):
        def get_jwks(self):
            return jwks

    validator = MyJWTBearerTokenValidator(
        issuer=issuer, resource_server=resource_server
    )
    return validator


@pytest.fixture(autouse=True)
def resource_protector(app, token_validator):
    require_oauth = ResourceProtector()
    require_oauth.register_token_validator(token_validator)

    @app.route("/protected")
    @require_oauth()
    def protected():
        user = db.session.get(User, current_token["sub"])
        return jsonify(
            id=user.id,
            username=user.username,
            token=current_token._get_current_object(),
        )

    @app.route("/protected-by-scope")
    @require_oauth("profile")
    def protected_by_scope():
        user = db.session.get(User, current_token["sub"])
        return jsonify(
            id=user.id,
            username=user.username,
            token=current_token._get_current_object(),
        )

    @app.route("/protected-by-groups")
    @require_oauth(groups=["admins"])
    def protected_by_groups():
        user = db.session.get(User, current_token["sub"])
        return jsonify(
            id=user.id,
            username=user.username,
            token=current_token._get_current_object(),
        )

    @app.route("/protected-by-roles")
    @require_oauth(roles=["student"])
    def protected_by_roles():
        user = db.session.get(User, current_token["sub"])
        return jsonify(
            id=user.id,
            username=user.username,
            token=current_token._get_current_object(),
        )

    @app.route("/protected-by-entitlements")
    @require_oauth(entitlements=["captain"])
    def protected_by_entitlements():
        user = db.session.get(User, current_token["sub"])
        return jsonify(
            id=user.id,
            username=user.username,
            token=current_token._get_current_object(),
        )

    return require_oauth


@pytest.fixture
def jwks():
    return read_file_path("jwks_private.json")


@pytest.fixture(autouse=True)
def user(db):
    user = User(username="foo")
    db.session.add(user)
    db.session.commit()
    yield user
    db.session.delete(user)


@pytest.fixture(autouse=True)
def client(client, db):
    client.set_client_metadata(
        {
            "scope": "profile",
            "redirect_uris": ["https://client.test/authorized"],
            "response_types": ["code"],
            "token_endpoint_auth_method": "client_secret_post",
            "grant_types": ["authorization_code"],
        }
    )
    db.session.add(client)
    db.session.commit()
    return client


def create_access_token_claims(client, user):
    now = int(time.time())
    expires_in = now + 3600
    auth_time = now - 60

    return {
        "iss": issuer,
        "exp": expires_in,
        "aud": resource_server,
        "sub": user.get_user_id(),
        "client_id": client.client_id,
        "iat": now,
        "jti": generate_token(16),
        "auth_time": auth_time,
        "scope": client.scope,
        "groups": ["admins"],
        "roles": ["student"],
        "entitlements": ["captain"],
    }


@pytest.fixture(autouse=True)
def claims(client, user):
    return create_access_token_claims(client, user)


def create_access_token(claims, jwks, alg="RS256", typ="at+jwt"):
    access_token = jwt.encode(
        {"alg": alg, "typ": typ},
        claims,
        key=jwks,
        check=False,
    )
    return access_token.decode()


@pytest.fixture
def access_token(claims, jwks):
    return create_access_token(claims, jwks)


@pytest.fixture
def token(access_token, user):
    token = Token(
        user_id=user.user_id,
        client_id="resource-server",
        token_type="bearer",
        access_token=access_token,
        scope="profile",
        expires_in=3600,
    )
    db.session.add(token)
    db.session.commit()
    yield token
    db.session.delete(token)


def test_access_resource(test_client, access_token):
    headers = {"Authorization": f"Bearer {access_token}"}

    rv = test_client.get("/protected", headers=headers)
    resp = json.loads(rv.data)
    assert resp["username"] == "foo"


def test_missing_authorization(test_client):
    rv = test_client.get("/protected")
    assert rv.status_code == 401
    resp = json.loads(rv.data)
    assert resp["error"] == "missing_authorization"


def test_unsupported_token_type(test_client):
    headers = {"Authorization": "invalid token"}
    rv = test_client.get("/protected", headers=headers)
    assert rv.status_code == 401
    resp = json.loads(rv.data)
    assert resp["error"] == "unsupported_token_type"


def test_invalid_token(test_client):
    headers = {"Authorization": "Bearer invalid"}
    rv = test_client.get("/protected", headers=headers)
    assert rv.status_code == 401
    resp = json.loads(rv.data)
    assert resp["error"] == "invalid_token"


def test_typ(test_client, access_token, claims, jwks):
    """The resource server MUST verify that the 'typ' header value is 'at+jwt' or
    'application/at+jwt' and reject tokens carrying any other value.
    """
    headers = {"Authorization": f"Bearer {access_token}"}
    rv = test_client.get("/protected", headers=headers)
    resp = json.loads(rv.data)
    assert resp["username"] == "foo"

    access_token = create_access_token(claims, jwks, typ="application/at+jwt")

    headers = {"Authorization": f"Bearer {access_token}"}
    rv = test_client.get("/protected", headers=headers)
    resp = json.loads(rv.data)
    assert resp["username"] == "foo"

    access_token = create_access_token(claims, jwks, typ="invalid")

    headers = {"Authorization": f"Bearer {access_token}"}
    rv = test_client.get("/protected", headers=headers)
    resp = json.loads(rv.data)
    assert resp["error"] == "invalid_token"


def test_missing_required_claims(test_client, client, user, jwks):
    required_claims = ["iss", "exp", "aud", "sub", "client_id", "iat", "jti"]
    for claim in required_claims:
        claims = create_access_token_claims(client, user)
        del claims[claim]
        access_token = create_access_token(claims, jwks)

        headers = {"Authorization": f"Bearer {access_token}"}
        rv = test_client.get("/protected", headers=headers)
        resp = json.loads(rv.data)
        assert resp["error"] == "invalid_token"


def test_invalid_iss(test_client, claims, jwks):
    """The issuer identifier for the authorization server (which is typically obtained
    during discovery) MUST exactly match the value of the 'iss' claim.
    """
    claims["iss"] = "invalid-issuer"
    access_token = create_access_token(claims, jwks)

    headers = {"Authorization": f"Bearer {access_token}"}
    rv = test_client.get("/protected", headers=headers)
    resp = json.loads(rv.data)
    assert resp["error"] == "invalid_token"


def test_invalid_aud(test_client, claims, jwks):
    """The resource server MUST validate that the 'aud' claim contains a resource
    indicator value corresponding to an identifier the resource server expects for
    itself. The JWT access token MUST be rejected if 'aud' does not contain a
    resource indicator of the current resource server as a valid audience.
    """
    claims["aud"] = "invalid-resource-indicator"
    access_token = create_access_token(claims, jwks)

    headers = {"Authorization": f"Bearer {access_token}"}
    rv = test_client.get("/protected", headers=headers)
    resp = json.loads(rv.data)
    assert resp["error"] == "invalid_token"


def test_invalid_exp(test_client, claims, jwks):
    """The current time MUST be before the time represented by the 'exp' claim.
    Implementers MAY provide for some small leeway, usually no more than a few
    minutes, to account for clock skew.
    """
    claims["exp"] = time.time() - 1
    access_token = create_access_token(claims, jwks)

    headers = {"Authorization": f"Bearer {access_token}"}
    rv = test_client.get("/protected", headers=headers)
    resp = json.loads(rv.data)
    assert resp["error"] == "invalid_token"


def test_scope_restriction(test_client, claims, jwks):
    """If an authorization request includes a scope parameter, the corresponding
    issued JWT access token SHOULD include a 'scope' claim as defined in Section
    4.2 of [RFC8693]. All the individual scope strings in the 'scope' claim MUST
    have meaning for the resources indicated in the 'aud' claim. See Section 5 for
    more considerations about the relationship between scope strings and resources
    indicated by the 'aud' claim.
    """
    claims["scope"] = ["invalid-scope"]
    access_token = create_access_token(claims, jwks)

    headers = {"Authorization": f"Bearer {access_token}"}
    rv = test_client.get("/protected", headers=headers)
    resp = json.loads(rv.data)
    assert resp["username"] == "foo"

    rv = test_client.get("/protected-by-scope", headers=headers)
    resp = json.loads(rv.data)
    assert resp["error"] == "insufficient_scope"


def test_entitlements_restriction(test_client, client, user, jwks):
    """Many authorization servers embed authorization attributes that go beyond the
    delegated scenarios described by [RFC7519] in the access tokens they issue.
    Typical examples include resource owner memberships in roles and groups that
    are relevant to the resource being accessed, entitlements assigned to the
    resource owner for the targeted resource that the authorization server knows
    about, and so on. An authorization server wanting to include such attributes
    in a JWT access token SHOULD use the 'groups', 'roles', and 'entitlements'
    attributes of the 'User' resource schema defined by Section 4.1.2 of
    [RFC7643]) as claim types.
    """
    for claim in ["groups", "roles", "entitlements"]:
        claims = create_access_token_claims(client, user)
        claims[claim] = ["invalid"]
        access_token = create_access_token(claims, jwks)

        headers = {"Authorization": f"Bearer {access_token}"}
        rv = test_client.get("/protected", headers=headers)
        resp = json.loads(rv.data)
        assert resp["username"] == "foo"

        rv = test_client.get(f"/protected-by-{claim}", headers=headers)
        resp = json.loads(rv.data)
        assert resp["error"] == "invalid_token"


def test_extra_attributes(test_client, claims, jwks):
    """Authorization servers MAY return arbitrary attributes not defined in any
    existing specification, as long as the corresponding claim names are collision
    resistant or the access tokens are meant to be used only within a private
    subsystem. Please refer to Sections 4.2 and 4.3 of [RFC7519] for details.
    """
    claims["email"] = "user@example.org"
    access_token = create_access_token(claims, jwks)

    headers = {"Authorization": f"Bearer {access_token}"}
    rv = test_client.get("/protected", headers=headers)
    resp = json.loads(rv.data)
    assert resp["token"]["email"] == "user@example.org"


def test_invalid_auth_time(test_client, claims, jwks):
    claims["auth_time"] = "invalid-auth-time"
    access_token = create_access_token(claims, jwks)

    headers = {"Authorization": f"Bearer {access_token}"}
    rv = test_client.get("/protected", headers=headers)
    resp = json.loads(rv.data)
    assert resp["error"] == "invalid_token"


def test_invalid_amr(test_client, claims, jwks):
    claims["amr"] = "invalid-amr"
    access_token = create_access_token(claims, jwks)

    headers = {"Authorization": f"Bearer {access_token}"}
    rv = test_client.get("/protected", headers=headers)
    resp = json.loads(rv.data)
    assert resp["error"] == "invalid_token"



"""authlib.oauth2.rfc6749.
~~~~~~~~~~~~~~~~~~~~~~

This module represents a direct implementation of
The OAuth 2.0 Authorization Framework.

https://tools.ietf.org/html/rfc6749
"""

from .authenticate_client import ClientAuthentication
from .authorization_server import AuthorizationServer
from .errors import AccessDeniedError
from .errors import InsecureTransportError
from .errors import InvalidClientError
from .errors import InvalidGrantError
from .errors import InvalidRequestError
from .errors import InvalidScopeError
from .errors import MismatchingStateException
from .errors import MissingAuthorizationError
from .errors import MissingCodeException  # exceptions for clients
from .errors import MissingTokenException
from .errors import MissingTokenTypeException
from .errors import OAuth2Error
from .errors import UnauthorizedClientError
from .errors import UnsupportedGrantTypeError
from .errors import UnsupportedResponseTypeError
from .errors import UnsupportedTokenTypeError
from .grants import AuthorizationCodeGrant
from .grants import AuthorizationEndpointMixin
from .grants import BaseGrant
from .grants import ClientCredentialsGrant
from .grants import ImplicitGrant
from .grants import RefreshTokenGrant
from .grants import ResourceOwnerPasswordCredentialsGrant
from .grants import TokenEndpointMixin
from .models import AuthorizationCodeMixin
from .models import ClientMixin
from .models import TokenMixin
from .requests import JsonPayload
from .requests import JsonRequest
from .requests import OAuth2Payload
from .requests import OAuth2Request
from .resource_protector import ResourceProtector
from .resource_protector import TokenValidator
from .token_endpoint import TokenEndpoint
from .util import list_to_scope
from .util import scope_to_list
from .wrappers import OAuth2Token

__all__ = [
    "OAuth2Payload",
    "OAuth2Token",
    "OAuth2Request",
    "JsonPayload",
    "JsonRequest",
    "OAuth2Error",
    "AccessDeniedError",
    "MissingAuthorizationError",
    "InvalidGrantError",
    "InvalidClientError",
    "InvalidRequestError",
    "InvalidScopeError",
    "InsecureTransportError",
    "UnauthorizedClientError",
    "UnsupportedResponseTypeError",
    "UnsupportedGrantTypeError",
    "UnsupportedTokenTypeError",
    "MissingCodeException",
    "MissingTokenException",
    "MissingTokenTypeException",
    "MismatchingStateException",
    "ClientMixin",
    "AuthorizationCodeMixin",
    "TokenMixin",
    "ClientAuthentication",
    "AuthorizationServer",
    "ResourceProtector",
    "TokenValidator",
    "TokenEndpoint",
    "BaseGrant",
    "AuthorizationEndpointMixin",
    "TokenEndpointMixin",
    "AuthorizationCodeGrant",
    "ImplicitGrant",
    "ResourceOwnerPasswordCredentialsGrant",
    "ClientCredentialsGrant",
    "RefreshTokenGrant",
    "scope_to_list",
    "list_to_scope",
]



import time
from copy import deepcopy
from unittest import mock

import pytest
from httpx import WSGITransport

from authlib.common.security import generate_token
from authlib.common.urls import url_encode
from authlib.integrations.httpx_client import OAuth2Client
from authlib.integrations.httpx_client import OAuthError

from ..wsgi_helper import MockDispatch

default_token = {
    "token_type": "Bearer",
    "access_token": "a",
    "refresh_token": "b",
    "expires_in": "3600",
    "expires_at": int(time.time()) + 3600,
}


def assert_token_in_header(request):
    token = "Bearer " + default_token["access_token"]
    auth_header = request.headers.get("authorization")
    assert auth_header == token


def assert_token_in_body(request):
    content = request.data
    content = content.decode()
    assert content == "access_token={}".format(default_token["access_token"])


def assert_token_in_uri(request):
    assert default_token["access_token"] in str(request.url)


@pytest.mark.parametrize(
    "assert_func, token_placement",
    [
        (assert_token_in_header, "header"),
        (assert_token_in_body, "body"),
        (assert_token_in_uri, "uri"),
    ],
)
def test_add_token_get_request(assert_func, token_placement):
    transport = WSGITransport(MockDispatch({"a": "a"}, assert_func=assert_func))
    with OAuth2Client(
        "foo", token=default_token, token_placement=token_placement, transport=transport
    ) as client:
        resp = client.get("https://provider.test")

    data = resp.json()
    assert data["a"] == "a"


@pytest.mark.parametrize(
    "assert_func, token_placement",
    [
        (assert_token_in_header, "header"),
        (assert_token_in_body, "body"),
        (assert_token_in_uri, "uri"),
    ],
)
def test_add_token_to_streaming_request(assert_func, token_placement):
    transport = WSGITransport(MockDispatch({"a": "a"}, assert_func=assert_func))
    with OAuth2Client(
        "foo", token=default_token, token_placement=token_placement, transport=transport
    ) as client:
        with client.stream("GET", "https://provider.test") as stream:
            stream.read()
            data = stream.json()
    assert data["a"] == "a"


def test_create_authorization_url():
    url = "https://provider.test/authorize?foo=bar"

    sess = OAuth2Client(client_id="foo")
    auth_url, state = sess.create_authorization_url(url)
    assert state in auth_url
    assert "client_id=foo" in auth_url
    assert "response_type=code" in auth_url

    sess = OAuth2Client(client_id="foo", prompt="none")
    auth_url, state = sess.create_authorization_url(
        url, state="foo", redirect_uri="https://provider.test", scope="profile"
    )
    assert state == "foo"
    assert "provider.test" in auth_url
    assert "profile" in auth_url
    assert "prompt=none" in auth_url


def test_code_challenge():
    sess = OAuth2Client("foo", code_challenge_method="S256")

    url = "https://provider.test/authorize"
    auth_url, _ = sess.create_authorization_url(url, code_verifier=generate_token(48))
    assert "code_challenge=" in auth_url
    assert "code_challenge_method=S256" in auth_url


def test_token_from_fragment():
    sess = OAuth2Client("foo")
    response_url = "https://provider.test/callback#" + url_encode(default_token.items())
    assert sess.token_from_fragment(response_url) == default_token
    token = sess.fetch_token(authorization_response=response_url)
    assert token == default_token


def test_fetch_token_post():
    url = "https://provider.test/token"

    def assert_func(request):
        content = request.form
        assert content.get("code") == "v"
        assert content.get("client_id") == "foo"
        assert content.get("grant_type") == "authorization_code"

    transport = WSGITransport(MockDispatch(default_token, assert_func=assert_func))
    with OAuth2Client("foo", transport=transport) as client:
        token = client.fetch_token(
            url, authorization_response="https://provider.test/?code=v"
        )
        assert token == default_token

    with OAuth2Client(
        "foo", token_endpoint_auth_method="none", transport=transport
    ) as client:
        token = client.fetch_token(url, code="v")
        assert token == default_token

    transport = WSGITransport(MockDispatch({"error": "invalid_request"}))
    with OAuth2Client("foo", transport=transport) as client:
        with pytest.raises(OAuthError):
            client.fetch_token(url)


def test_fetch_token_get():
    url = "https://provider.test/token"

    def assert_func(request):
        url = str(request.url)
        assert "code=v" in url
        assert "client_id=" in url
        assert "grant_type=authorization_code" in url

    transport = WSGITransport(MockDispatch(default_token, assert_func=assert_func))
    with OAuth2Client("foo", transport=transport) as client:
        authorization_response = "https://provider.test/?code=v"
        token = client.fetch_token(
            url, authorization_response=authorization_response, method="GET"
        )
        assert token == default_token

    with OAuth2Client(
        "foo", token_endpoint_auth_method="none", transport=transport
    ) as client:
        token = client.fetch_token(url, code="v", method="GET")
        assert token == default_token

        token = client.fetch_token(url + "?q=a", code="v", method="GET")
        assert token == default_token


def test_token_auth_method_client_secret_post():
    url = "https://provider.test/token"

    def assert_func(request):
        content = request.form
        assert content.get("code") == "v"
        assert content.get("client_id") == "foo"
        assert content.get("client_secret") == "bar"
        assert content.get("grant_type") == "authorization_code"

    transport = WSGITransport(MockDispatch(default_token, assert_func=assert_func))
    with OAuth2Client(
        "foo",
        "bar",
        token_endpoint_auth_method="client_secret_post",
        transport=transport,
    ) as client:
        token = client.fetch_token(url, code="v")

    assert token == default_token


def test_access_token_response_hook():
    url = "https://provider.test/token"

    def _access_token_response_hook(resp):
        assert resp.json() == default_token
        return resp

    access_token_response_hook = mock.Mock(side_effect=_access_token_response_hook)
    transport = WSGITransport(MockDispatch(default_token))
    with OAuth2Client("foo", token=default_token, transport=transport) as sess:
        sess.register_compliance_hook(
            "access_token_response", access_token_response_hook
        )
        assert sess.fetch_token(url) == default_token
        assert access_token_response_hook.called is True


def test_password_grant_type():
    url = "https://provider.test/token"

    def assert_func(request):
        content = request.form
        assert content.get("username") == "v"
        assert content.get("scope") == "profile"
        assert content.get("grant_type") == "password"

    transport = WSGITransport(MockDispatch(default_token, assert_func=assert_func))
    with OAuth2Client("foo", scope="profile", transport=transport) as sess:
        token = sess.fetch_token(url, username="v", password="v")
        assert token == default_token

        token = sess.fetch_token(url, username="v", password="v", grant_type="password")
        assert token == default_token


def test_client_credentials_type():
    url = "https://provider.test/token"

    def assert_func(request):
        content = request.form
        assert content.get("scope") == "profile"
        assert content.get("grant_type") == "client_credentials"

    transport = WSGITransport(MockDispatch(default_token, assert_func=assert_func))
    with OAuth2Client("foo", scope="profile", transport=transport) as sess:
        token = sess.fetch_token(url)
        assert token == default_token

        token = sess.fetch_token(url, grant_type="client_credentials")
        assert token == default_token


def test_cleans_previous_token_before_fetching_new_one():
    now = int(time.time())
    new_token = deepcopy(default_token)
    past = now - 7200
    default_token["expires_at"] = past
    new_token["expires_at"] = now + 3600
    url = "https://provider.test/token"

    transport = WSGITransport(MockDispatch(new_token))
    with mock.patch("time.time", lambda: now):
        with OAuth2Client("foo", token=default_token, transport=transport) as sess:
            assert sess.fetch_token(url) == new_token


def test_token_status():
    token = dict(access_token="a", token_type="bearer", expires_at=100)
    sess = OAuth2Client("foo", token=token)
    assert sess.token.is_expired() is True


def test_auto_refresh_token():
    def _update_token(token, refresh_token=None, access_token=None):
        assert refresh_token == "b"
        assert token == default_token

    update_token = mock.Mock(side_effect=_update_token)

    old_token = dict(
        access_token="a", refresh_token="b", token_type="bearer", expires_at=100
    )

    transport = WSGITransport(MockDispatch(default_token))
    with OAuth2Client(
        "foo",
        token=old_token,
        token_endpoint="https://provider.test/token",
        update_token=update_token,
        transport=transport,
    ) as sess:
        sess.get("https://resource.test/user")
        assert update_token.called is True

    old_token = dict(access_token="a", token_type="bearer", expires_at=100)
    with OAuth2Client(
        "foo",
        token=old_token,
        token_endpoint="https://provider.test/token",
        update_token=update_token,
        transport=transport,
    ) as sess:
        with pytest.raises(OAuthError):
            sess.get("https://resource.test/user")


def test_auto_refresh_token2():
    def _update_token(token, refresh_token=None, access_token=None):
        assert access_token == "a"
        assert token == default_token

    update_token = mock.Mock(side_effect=_update_token)

    old_token = dict(access_token="a", token_type="bearer", expires_at=100)

    transport = WSGITransport(MockDispatch(default_token))

    with OAuth2Client(
        "foo",
        token=old_token,
        token_endpoint="https://provider.test/token",
        grant_type="client_credentials",
        transport=transport,
    ) as client:
        client.get("https://resource.test/user")
        assert update_token.called is False

    with OAuth2Client(
        "foo",
        token=old_token,
        token_endpoint="https://provider.test/token",
        update_token=update_token,
        grant_type="client_credentials",
        transport=transport,
    ) as client:
        client.get("https://resource.test/user")
        assert update_token.called is True


def test_auto_refresh_token3():
    def _update_token(token, refresh_token=None, access_token=None):
        assert access_token == "a"
        assert token == default_token

    update_token = mock.Mock(side_effect=_update_token)

    old_token = dict(access_token="a", token_type="bearer", expires_at=100)

    transport = WSGITransport(MockDispatch(default_token))

    with OAuth2Client(
        "foo",
        token=old_token,
        token_endpoint="https://provider.test/token",
        update_token=update_token,
        grant_type="client_credentials",
        transport=transport,
    ) as client:
        client.post("https://resource.test/user", json={"foo": "bar"})
        assert update_token.called is True


def test_revoke_token():
    answer = {"status": "ok"}
    transport = WSGITransport(MockDispatch(answer))

    with OAuth2Client("a", transport=transport) as sess:
        resp = sess.revoke_token("https://provider.test/token", "hi")
        assert resp.json() == answer

        resp = sess.revoke_token(
            "https://provider.test/token", "hi", token_type_hint="access_token"
        )
        assert resp.json() == answer


def test_request_without_token():
    transport = WSGITransport(MockDispatch())
    with OAuth2Client("a", transport=transport) as client:
        with pytest.raises(OAuthError):
            client.get("https://provider.test/token")



from cryptography.hazmat.primitives.asymmetric.ed448 import Ed448PrivateKey
from cryptography.hazmat.primitives.asymmetric.ed448 import Ed448PublicKey
from cryptography.hazmat.primitives.asymmetric.ed25519 import Ed25519PrivateKey
from cryptography.hazmat.primitives.asymmetric.ed25519 import Ed25519PublicKey
from cryptography.hazmat.primitives.asymmetric.x448 import X448PrivateKey
from cryptography.hazmat.primitives.asymmetric.x448 import X448PublicKey
from cryptography.hazmat.primitives.asymmetric.x25519 import X25519PrivateKey
from cryptography.hazmat.primitives.asymmetric.x25519 import X25519PublicKey
from cryptography.hazmat.primitives.serialization import Encoding
from cryptography.hazmat.primitives.serialization import NoEncryption
from cryptography.hazmat.primitives.serialization import PrivateFormat
from cryptography.hazmat.primitives.serialization import PublicFormat

from authlib.common.encoding import to_bytes
from authlib.common.encoding import to_unicode
from authlib.common.encoding import urlsafe_b64decode
from authlib.common.encoding import urlsafe_b64encode

from ..rfc7517 import AsymmetricKey

PUBLIC_KEYS_MAP = {
    "Ed25519": Ed25519PublicKey,
    "Ed448": Ed448PublicKey,
    "X25519": X25519PublicKey,
    "X448": X448PublicKey,
}
PRIVATE_KEYS_MAP = {
    "Ed25519": Ed25519PrivateKey,
    "Ed448": Ed448PrivateKey,
    "X25519": X25519PrivateKey,
    "X448": X448PrivateKey,
}


class OKPKey(AsymmetricKey):
    """Key class of the ``OKP`` key type."""

    kty = "OKP"
    REQUIRED_JSON_FIELDS = ["crv", "x"]
    PUBLIC_KEY_FIELDS = REQUIRED_JSON_FIELDS
    PRIVATE_KEY_FIELDS = ["crv", "d"]
    PUBLIC_KEY_CLS = tuple(PUBLIC_KEYS_MAP.values())
    PRIVATE_KEY_CLS = tuple(PRIVATE_KEYS_MAP.values())
    SSH_PUBLIC_PREFIX = b"ssh-ed25519"

    def exchange_shared_key(self, pubkey):
        # used in ECDHESAlgorithm
        private_key = self.get_private_key()
        if private_key and isinstance(private_key, (X25519PrivateKey, X448PrivateKey)):
            return private_key.exchange(pubkey)
        raise ValueError("Invalid key for exchanging shared key")

    @staticmethod
    def get_key_curve(key):
        if isinstance(key, (Ed25519PublicKey, Ed25519PrivateKey)):
            return "Ed25519"
        elif isinstance(key, (Ed448PublicKey, Ed448PrivateKey)):
            return "Ed448"
        elif isinstance(key, (X25519PublicKey, X25519PrivateKey)):
            return "X25519"
        elif isinstance(key, (X448PublicKey, X448PrivateKey)):
            return "X448"

    def load_private_key(self):
        crv_key = PRIVATE_KEYS_MAP[self._dict_data["crv"]]
        d_bytes = urlsafe_b64decode(to_bytes(self._dict_data["d"]))
        return crv_key.from_private_bytes(d_bytes)

    def load_public_key(self):
        crv_key = PUBLIC_KEYS_MAP[self._dict_data["crv"]]
        x_bytes = urlsafe_b64decode(to_bytes(self._dict_data["x"]))
        return crv_key.from_public_bytes(x_bytes)

    def dumps_private_key(self):
        obj = self.dumps_public_key(self.private_key.public_key())
        d_bytes = self.private_key.private_bytes(
            Encoding.Raw, PrivateFormat.Raw, NoEncryption()
        )
        obj["d"] = to_unicode(urlsafe_b64encode(d_bytes))
        return obj

    def dumps_public_key(self, public_key=None):
        if public_key is None:
            public_key = self.public_key
        x_bytes = public_key.public_bytes(Encoding.Raw, PublicFormat.Raw)
        return {
            "crv": self.get_key_curve(public_key),
            "x": to_unicode(urlsafe_b64encode(x_bytes)),
        }

    @classmethod
    def generate_key(cls, crv="Ed25519", options=None, is_private=False) -> "OKPKey":
        if crv not in PRIVATE_KEYS_MAP:
            raise ValueError(f'Invalid crv value: "{crv}"')
        private_key_cls = PRIVATE_KEYS_MAP[crv]
        raw_key = private_key_cls.generate()
        if not is_private:
            raw_key = raw_key.public_key()
        return cls.import_key(raw_key, options=options)



import pytest
from flask import json

from authlib.oauth2.rfc7591 import ClientMetadataClaims as OAuth2ClientMetadataClaims
from authlib.oauth2.rfc7591 import (
    ClientRegistrationEndpoint as _ClientRegistrationEndpoint,
)
from authlib.oidc.registration import ClientMetadataClaims as OIDCClientMetadataClaims
from tests.util import read_file_path

from .models import Client
from .models import db


class ClientRegistrationEndpoint(_ClientRegistrationEndpoint):
    software_statement_alg_values_supported = ["RS256"]

    def authenticate_token(self, request):
        auth_header = request.headers.get("Authorization")
        if auth_header:
            request.user_id = 1
            return auth_header

    def resolve_public_key(self, request):
        return read_file_path("rsa_public.pem")

    def save_client(self, client_info, client_metadata, request):
        client = Client(user_id=request.user_id, **client_info)
        client.set_client_metadata(client_metadata)
        db.session.add(client)
        db.session.commit()
        return client


@pytest.fixture
def metadata():
    return {}


@pytest.fixture(autouse=True)
def server(server, app, metadata):
    class MyClientRegistration(ClientRegistrationEndpoint):
        def get_server_metadata(self):
            return metadata

    server.register_endpoint(
        MyClientRegistration(
            claims_classes=[OAuth2ClientMetadataClaims, OIDCClientMetadataClaims]
        )
    )

    @app.route("/create_client", methods=["POST"])
    def create_client():
        return server.create_endpoint_response("client_registration")

    return server


def test_application_type(test_client):
    # Nominal case
    body = {
        "application_type": "web",
        "client_name": "Authlib",
    }
    rv = test_client.post(
        "/create_client", json=body, headers={"Authorization": "bearer abc"}
    )
    resp = json.loads(rv.data)
    assert "client_id" in resp
    assert resp["client_name"] == "Authlib"
    assert resp["application_type"] == "web"

    # Default case
    # The default, if omitted, is that any algorithm supported by the OP and the RP MAY be used.
    body = {
        "client_name": "Authlib",
    }
    rv = test_client.post(
        "/create_client", json=body, headers={"Authorization": "bearer abc"}
    )
    resp = json.loads(rv.data)
    assert "client_id" in resp
    assert resp["client_name"] == "Authlib"
    assert resp["application_type"] == "web"

    # Error case
    body = {
        "application_type": "invalid",
        "client_name": "Authlib",
    }
    rv = test_client.post(
        "/create_client", json=body, headers={"Authorization": "bearer abc"}
    )
    resp = json.loads(rv.data)
    assert resp["error"] in "invalid_client_metadata"


def test_token_endpoint_auth_signing_alg_supported(test_client, metadata):
    metadata["token_endpoint_auth_signing_alg_values_supported"] = ["RS256", "ES256"]

    # Nominal case
    body = {
        "token_endpoint_auth_signing_alg": "ES256",
        "client_name": "Authlib",
    }
    rv = test_client.post(
        "/create_client", json=body, headers={"Authorization": "bearer abc"}
    )
    resp = json.loads(rv.data)
    assert "client_id" in resp
    assert resp["client_name"] == "Authlib"
    assert resp["token_endpoint_auth_signing_alg"] == "ES256"

    # Default case
    # The default, if omitted, is that any algorithm supported by the OP and the RP MAY be used.
    body = {
        "client_name": "Authlib",
    }
    rv = test_client.post(
        "/create_client", json=body, headers={"Authorization": "bearer abc"}
    )
    resp = json.loads(rv.data)
    assert "client_id" in resp
    assert resp["client_name"] == "Authlib"

    # Error case
    body = {
        "token_endpoint_auth_signing_alg": "RS512",
        "client_name": "Authlib",
    }
    rv = test_client.post(
        "/create_client", json=body, headers={"Authorization": "bearer abc"}
    )
    resp = json.loads(rv.data)
    assert resp["error"] in "invalid_client_metadata"


def test_subject_types_supported(test_client, metadata):
    metadata["subject_types_supported"] = ["public", "pairwise"]

    # Nominal case
    body = {"subject_type": "public", "client_name": "Authlib"}
    rv = test_client.post(
        "/create_client", json=body, headers={"Authorization": "bearer abc"}
    )
    resp = json.loads(rv.data)
    assert "client_id" in resp
    assert resp["client_name"] == "Authlib"
    assert resp["subject_type"] == "public"

    # Error case
    body = {"subject_type": "invalid", "client_name": "Authlib"}
    rv = test_client.post(
        "/create_client", json=body, headers={"Authorization": "bearer abc"}
    )
    resp = json.loads(rv.data)
    assert resp["error"] in "invalid_client_metadata"


def test_id_token_signing_alg_values_supported(test_client, metadata):
    metadata["id_token_signing_alg_values_supported"] = ["RS256", "ES256"]

    # Default
    # The default, if omitted, is RS256.
    body = {"client_name": "Authlib"}
    rv = test_client.post(
        "/create_client", json=body, headers={"Authorization": "bearer abc"}
    )
    resp = json.loads(rv.data)
    assert "client_id" in resp
    assert resp["client_name"] == "Authlib"
    assert resp["id_token_signed_response_alg"] == "RS256"

    # Nominal case
    body = {"id_token_signed_response_alg": "ES256", "client_name": "Authlib"}
    rv = test_client.post(
        "/create_client", json=body, headers={"Authorization": "bearer abc"}
    )
    resp = json.loads(rv.data)
    assert "client_id" in resp
    assert resp["client_name"] == "Authlib"
    assert resp["id_token_signed_response_alg"] == "ES256"

    # Error case
    body = {"id_token_signed_response_alg": "RS512", "client_name": "Authlib"}
    rv = test_client.post(
        "/create_client", json=body, headers={"Authorization": "bearer abc"}
    )
    resp = json.loads(rv.data)
    assert resp["error"] == "invalid_client_metadata"


def test_id_token_signing_alg_values_none(test_client, metadata):
    # The value none MUST NOT be used as the ID Token alg value unless the Client uses
    # only Response Types that return no ID Token from the Authorization Endpoint
    # (such as when only using the Authorization Code Flow).
    metadata["id_token_signing_alg_values_supported"] = ["none", "RS256", "ES256"]

    # Nominal case
    body = {
        "id_token_signed_response_alg": "none",
        "client_name": "Authlib",
        "response_type": "code",
    }
    rv = test_client.post(
        "/create_client", json=body, headers={"Authorization": "bearer abc"}
    )
    resp = json.loads(rv.data)
    assert "client_id" in resp
    assert resp["client_name"] == "Authlib"
    assert resp["id_token_signed_response_alg"] == "none"

    # Error case
    body = {
        "id_token_signed_response_alg": "none",
        "client_name": "Authlib",
        "response_type": "id_token",
    }
    rv = test_client.post(
        "/create_client", json=body, headers={"Authorization": "bearer abc"}
    )
    resp = json.loads(rv.data)
    assert resp["error"] == "invalid_client_metadata"


def test_id_token_encryption_alg_values_supported(test_client, metadata):
    metadata["id_token_encryption_alg_values_supported"] = ["RS256", "ES256"]

    # Default case
    body = {"client_name": "Authlib"}
    rv = test_client.post(
        "/create_client", json=body, headers={"Authorization": "bearer abc"}
    )
    resp = json.loads(rv.data)
    assert "client_id" in resp
    assert resp["client_name"] == "Authlib"
    assert "id_token_encrypted_response_enc" not in resp

    # If id_token_encrypted_response_alg is specified, the default
    # id_token_encrypted_response_enc value is A128CBC-HS256.
    body = {"id_token_encrypted_response_alg": "RS256", "client_name": "Authlib"}
    rv = test_client.post(
        "/create_client", json=body, headers={"Authorization": "bearer abc"}
    )
    resp = json.loads(rv.data)
    assert "client_id" in resp
    assert resp["client_name"] == "Authlib"
    assert resp["id_token_encrypted_response_enc"] == "A128CBC-HS256"

    # Nominal case
    body = {"id_token_encrypted_response_alg": "ES256", "client_name": "Authlib"}
    rv = test_client.post(
        "/create_client", json=body, headers={"Authorization": "bearer abc"}
    )
    resp = json.loads(rv.data)
    assert "client_id" in resp
    assert resp["client_name"] == "Authlib"
    assert resp["id_token_encrypted_response_alg"] == "ES256"

    # Error case
    body = {"id_token_encrypted_response_alg": "RS512", "client_name": "Authlib"}
    rv = test_client.post(
        "/create_client", json=body, headers={"Authorization": "bearer abc"}
    )
    resp = json.loads(rv.data)
    assert resp["error"] in "invalid_client_metadata"


def test_id_token_encryption_enc_values_supported(test_client, metadata):
    metadata["id_token_encryption_enc_values_supported"] = ["A128CBC-HS256", "A256GCM"]

    # Nominal case
    body = {
        "id_token_encrypted_response_alg": "RS256",
        "id_token_encrypted_response_enc": "A256GCM",
        "client_name": "Authlib",
    }
    rv = test_client.post(
        "/create_client", json=body, headers={"Authorization": "bearer abc"}
    )
    resp = json.loads(rv.data)
    assert "client_id" in resp
    assert resp["client_name"] == "Authlib"
    assert resp["id_token_encrypted_response_alg"] == "RS256"
    assert resp["id_token_encrypted_response_enc"] == "A256GCM"

    # Error case: missing id_token_encrypted_response_alg
    body = {"id_token_encrypted_response_enc": "A256GCM", "client_name": "Authlib"}
    rv = test_client.post(
        "/create_client", json=body, headers={"Authorization": "bearer abc"}
    )
    resp = json.loads(rv.data)
    assert resp["error"] in "invalid_client_metadata"

    # Error case: alg not in server metadata
    body = {"id_token_encrypted_response_enc": "A128GCM", "client_name": "Authlib"}
    rv = test_client.post(
        "/create_client", json=body, headers={"Authorization": "bearer abc"}
    )
    resp = json.loads(rv.data)
    assert resp["error"] in "invalid_client_metadata"


def test_userinfo_signing_alg_values_supported(test_client, metadata):
    metadata["userinfo_signing_alg_values_supported"] = ["RS256", "ES256"]

    # Nominal case
    body = {"userinfo_signed_response_alg": "ES256", "client_name": "Authlib"}
    rv = test_client.post(
        "/create_client", json=body, headers={"Authorization": "bearer abc"}
    )
    resp = json.loads(rv.data)
    assert "client_id" in resp
    assert resp["client_name"] == "Authlib"
    assert resp["userinfo_signed_response_alg"] == "ES256"

    # Error case
    body = {"userinfo_signed_response_alg": "RS512", "client_name": "Authlib"}
    rv = test_client.post(
        "/create_client", json=body, headers={"Authorization": "bearer abc"}
    )
    resp = json.loads(rv.data)
    assert resp["error"] in "invalid_client_metadata"


def test_userinfo_encryption_alg_values_supported(test_client, metadata):
    metadata["userinfo_encryption_alg_values_supported"] = ["RS256", "ES256"]

    # Nominal case
    body = {"userinfo_encrypted_response_alg": "ES256", "client_name": "Authlib"}
    rv = test_client.post(
        "/create_client", json=body, headers={"Authorization": "bearer abc"}
    )
    resp = json.loads(rv.data)
    assert "client_id" in resp
    assert resp["client_name"] == "Authlib"
    assert resp["userinfo_encrypted_response_alg"] == "ES256"

    # Error case
    body = {"userinfo_encrypted_response_alg": "RS512", "client_name": "Authlib"}
    rv = test_client.post(
        "/create_client", json=body, headers={"Authorization": "bearer abc"}
    )
    resp = json.loads(rv.data)
    assert resp["error"] in "invalid_client_metadata"


def test_userinfo_encryption_enc_values_supported(test_client, metadata):
    metadata["userinfo_encryption_enc_values_supported"] = ["A128CBC-HS256", "A256GCM"]

    # Default case
    body = {"client_name": "Authlib"}
    rv = test_client.post(
        "/create_client", json=body, headers={"Authorization": "bearer abc"}
    )
    resp = json.loads(rv.data)
    assert "client_id" in resp
    assert resp["client_name"] == "Authlib"
    assert "userinfo_encrypted_response_enc" not in resp

    # If userinfo_encrypted_response_alg is specified, the default
    # userinfo_encrypted_response_enc value is A128CBC-HS256.
    body = {"userinfo_encrypted_response_alg": "RS256", "client_name": "Authlib"}
    rv = test_client.post(
        "/create_client", json=body, headers={"Authorization": "bearer abc"}
    )
    resp = json.loads(rv.data)
    assert "client_id" in resp
    assert resp["client_name"] == "Authlib"
    assert resp["userinfo_encrypted_response_enc"] == "A128CBC-HS256"

    # Nominal case
    body = {
        "userinfo_encrypted_response_alg": "RS256",
        "userinfo_encrypted_response_enc": "A256GCM",
        "client_name": "Authlib",
    }
    rv = test_client.post(
        "/create_client", json=body, headers={"Authorization": "bearer abc"}
    )
    resp = json.loads(rv.data)
    assert "client_id" in resp
    assert resp["client_name"] == "Authlib"
    assert resp["userinfo_encrypted_response_alg"] == "RS256"
    assert resp["userinfo_encrypted_response_enc"] == "A256GCM"

    # Error case: no userinfo_encrypted_response_alg
    body = {"userinfo_encrypted_response_enc": "A256GCM", "client_name": "Authlib"}
    rv = test_client.post(
        "/create_client", json=body, headers={"Authorization": "bearer abc"}
    )
    resp = json.loads(rv.data)
    assert resp["error"] in "invalid_client_metadata"

    # Error case: alg not in server metadata
    body = {"userinfo_encrypted_response_enc": "A128GCM", "client_name": "Authlib"}
    rv = test_client.post(
        "/create_client", json=body, headers={"Authorization": "bearer abc"}
    )
    resp = json.loads(rv.data)
    assert resp["error"] in "invalid_client_metadata"


def test_acr_values_supported(test_client, metadata):
    metadata["acr_values_supported"] = [
        "urn:mace:incommon:iap:silver",
        "urn:mace:incommon:iap:bronze",
    ]

    # Nominal case
    body = {
        "default_acr_values": ["urn:mace:incommon:iap:silver"],
        "client_name": "Authlib",
    }
    rv = test_client.post(
        "/create_client", json=body, headers={"Authorization": "bearer abc"}
    )
    resp = json.loads(rv.data)
    assert "client_id" in resp
    assert resp["client_name"] == "Authlib"
    assert resp["default_acr_values"] == ["urn:mace:incommon:iap:silver"]

    # Error case
    body = {
        "default_acr_values": [
            "urn:mace:incommon:iap:silver",
            "urn:mace:incommon:iap:gold",
        ],
        "client_name": "Authlib",
    }
    rv = test_client.post(
        "/create_client", json=body, headers={"Authorization": "bearer abc"}
    )
    resp = json.loads(rv.data)
    assert resp["error"] in "invalid_client_metadata"


def test_request_object_signing_alg_values_supported(test_client, metadata):
    metadata["request_object_signing_alg_values_supported"] = ["RS256", "ES256"]

    # Nominal case
    body = {"request_object_signing_alg": "ES256", "client_name": "Authlib"}
    rv = test_client.post(
        "/create_client", json=body, headers={"Authorization": "bearer abc"}
    )
    resp = json.loads(rv.data)
    assert "client_id" in resp
    assert resp["client_name"] == "Authlib"
    assert resp["request_object_signing_alg"] == "ES256"

    # Error case
    body = {"request_object_signing_alg": "RS512", "client_name": "Authlib"}
    rv = test_client.post(
        "/create_client", json=body, headers={"Authorization": "bearer abc"}
    )
    resp = json.loads(rv.data)
    assert resp["error"] in "invalid_client_metadata"


def test_request_object_encryption_alg_values_supported(test_client, metadata):
    metadata["request_object_encryption_alg_values_supported"] = ["RS256", "ES256"]

    # Nominal case
    body = {
        "request_object_encryption_alg": "ES256",
        "client_name": "Authlib",
    }
    rv = test_client.post(
        "/create_client", json=body, headers={"Authorization": "bearer abc"}
    )
    resp = json.loads(rv.data)
    assert "client_id" in resp
    assert resp["client_name"] == "Authlib"
    assert resp["request_object_encryption_alg"] == "ES256"

    # Error case
    body = {
        "request_object_encryption_alg": "RS512",
        "client_name": "Authlib",
    }
    rv = test_client.post(
        "/create_client", json=body, headers={"Authorization": "bearer abc"}
    )
    resp = json.loads(rv.data)
    assert resp["error"] in "invalid_client_metadata"


def test_request_object_encryption_enc_values_supported(test_client, metadata):
    metadata["request_object_encryption_enc_values_supported"] = [
        "A128CBC-HS256",
        "A256GCM",
    ]

    # Default case
    body = {"client_name": "Authlib"}
    rv = test_client.post(
        "/create_client", json=body, headers={"Authorization": "bearer abc"}
    )
    resp = json.loads(rv.data)
    assert "client_id" in resp
    assert resp["client_name"] == "Authlib"
    assert "request_object_encryption_enc" not in resp

    # If request_object_encryption_alg is specified, the default
    # request_object_encryption_enc value is A128CBC-HS256.
    body = {"request_object_encryption_alg": "RS256", "client_name": "Authlib"}
    rv = test_client.post(
        "/create_client", json=body, headers={"Authorization": "bearer abc"}
    )
    resp = json.loads(rv.data)
    assert "client_id" in resp
    assert resp["client_name"] == "Authlib"
    assert resp["request_object_encryption_enc"] == "A128CBC-HS256"

    # Nominal case
    body = {
        "request_object_encryption_alg": "RS256",
        "request_object_encryption_enc": "A256GCM",
        "client_name": "Authlib",
    }
    rv = test_client.post(
        "/create_client", json=body, headers={"Authorization": "bearer abc"}
    )
    resp = json.loads(rv.data)
    assert "client_id" in resp
    assert resp["client_name"] == "Authlib"
    assert resp["request_object_encryption_alg"] == "RS256"
    assert resp["request_object_encryption_enc"] == "A256GCM"

    # Error case: missing request_object_encryption_alg
    body = {
        "request_object_encryption_enc": "A256GCM",
        "client_name": "Authlib",
    }
    rv = test_client.post(
        "/create_client", json=body, headers={"Authorization": "bearer abc"}
    )
    resp = json.loads(rv.data)
    assert resp["error"] in "invalid_client_metadata"

    # Error case: alg not in server metadata
    body = {
        "request_object_encryption_enc": "A128GCM",
        "client_name": "Authlib",
    }
    rv = test_client.post(
        "/create_client", json=body, headers={"Authorization": "bearer abc"}
    )
    resp = json.loads(rv.data)
    assert resp["error"] in "invalid_client_metadata"


def test_require_auth_time(test_client):
    # Default case
    body = {
        "client_name": "Authlib",
    }
    rv = test_client.post(
        "/create_client", json=body, headers={"Authorization": "bearer abc"}
    )
    resp = json.loads(rv.data)
    assert "client_id" in resp
    assert resp["client_name"] == "Authlib"
    assert resp["require_auth_time"] is False

    # Nominal case
    body = {
        "require_auth_time": True,
        "client_name": "Authlib",
    }
    rv = test_client.post(
        "/create_client", json=body, headers={"Authorization": "bearer abc"}
    )
    resp = json.loads(rv.data)
    assert "client_id" in resp
    assert resp["client_name"] == "Authlib"
    assert resp["require_auth_time"] is True

    # Error case
    body = {
        "require_auth_time": "invalid",
        "client_name": "Authlib",
    }
    rv = test_client.post(
        "/create_client", json=body, headers={"Authorization": "bearer abc"}
    )
    resp = json.loads(rv.data)
    assert resp["error"] in "invalid_client_metadata"


def test_redirect_uri(test_client):
    """RFC6749 indicate that fragments are forbidden in redirect_uri.

        The redirection endpoint URI MUST be an absolute URI as defined by
        [RFC3986] Section 4.3.  [...]  The endpoint URI MUST NOT include a
        fragment component.

    https://www.rfc-editor.org/rfc/rfc6749#section-3.1.2
    """
    # Nominal case
    body = {
        "redirect_uris": ["https://client.test"],
        "client_name": "Authlib",
    }
    rv = test_client.post(
        "/create_client", json=body, headers={"Authorization": "bearer abc"}
    )
    resp = json.loads(rv.data)
    assert "client_id" in resp
    assert resp["client_name"] == "Authlib"
    assert resp["redirect_uris"] == ["https://client.test"]

    # Error case
    body = {
        "redirect_uris": ["https://client.test#fragment"],
        "client_name": "Authlib",
    }
    rv = test_client.post(
        "/create_client", json=body, headers={"Authorization": "bearer abc"}
    )
    resp = json.loads(rv.data)
    assert resp["error"] in "invalid_client_metadata"



import pytest
from flask import json

import authlib.oidc.core as oidc_core
from authlib.integrations.flask_oauth2 import ResourceProtector
from authlib.integrations.sqla_oauth2 import create_bearer_token_validator
from authlib.jose import JsonWebToken
from authlib.jose import jwt
from tests.util import read_file_path

from .models import Token


@pytest.fixture(autouse=True)
def server(server, app, db):
    class UserInfoEndpoint(oidc_core.UserInfoEndpoint):
        def get_issuer(self) -> str:
            return "https://provider.test"

        def generate_user_info(self, user, scope):
            return user.generate_user_info().filter(scope)

        def resolve_private_key(self):
            return read_file_path("jwks_private.json")

    BearerTokenValidator = create_bearer_token_validator(db.session, Token)
    resource_protector = ResourceProtector()
    resource_protector.register_token_validator(BearerTokenValidator())
    server.register_endpoint(UserInfoEndpoint(resource_protector=resource_protector))

    @app.route("/oauth/userinfo", methods=["GET", "POST"])
    def userinfo():
        return server.create_endpoint_response("userinfo")

    return server


@pytest.fixture(autouse=True)
def client(client, db):
    client.set_client_metadata(
        {
            "scope": "profile",
            "redirect_uris": ["https://client.test/authorized"],
        }
    )
    db.session.add(client)
    db.session.commit()
    return client


@pytest.fixture(autouse=True)
def token(db):
    token = Token(
        user_id=1,
        client_id="client-id",
        token_type="bearer",
        access_token="access-token",
        refresh_token="r1",
        scope="openid",
        expires_in=3600,
    )
    db.session.add(token)
    db.session.commit()
    yield token
    db.session.delete(token)


def test_get(test_client, db, token):
    """The UserInfo Endpoint MUST support the use of the HTTP GET and HTTP POST methods defined in RFC 7231 [RFC7231].
    The UserInfo Endpoint MUST accept Access Tokens as OAuth 2.0 Bearer Token Usage [RFC6750]."""

    token.scope = "openid profile email address phone"
    db.session.add(token)
    db.session.commit()

    headers = {"Authorization": "Bearer access-token"}
    rv = test_client.get("/oauth/userinfo", headers=headers)
    assert rv.headers["Content-Type"] == "application/json"

    resp = json.loads(rv.data)
    assert resp == {
        "sub": "1",
        "address": {
            "country": "USA",
            "formatted": "742 Evergreen Terrace, Springfield",
            "locality": "Springfield",
            "postal_code": "1245",
            "region": "Unknown",
            "street_address": "742 Evergreen Terrace",
        },
        "birthdate": "2000-12-01",
        "email": "janedoe@example.com",
        "email_verified": True,
        "family_name": "Doe",
        "gender": "female",
        "given_name": "Jane",
        "locale": "fr-FR",
        "middle_name": "Middle",
        "name": "foo",
        "nickname": "Jany",
        "phone_number": "+1 (425) 555-1212",
        "phone_number_verified": False,
        "picture": "https://resource.test/janedoe/me.jpg",
        "preferred_username": "j.doe",
        "profile": "https://resource.test/janedoe",
        "updated_at": 1745315119,
        "website": "https://resource.test",
        "zoneinfo": "Europe/Paris",
    }


def test_post(test_client, db, token):
    """The UserInfo Endpoint MUST support the use of the HTTP GET and HTTP POST methods defined in RFC 7231 [RFC7231].
    The UserInfo Endpoint MUST accept Access Tokens as OAuth 2.0 Bearer Token Usage [RFC6750]."""

    token.scope = "openid profile email address phone"
    db.session.add(token)
    db.session.commit()

    headers = {"Authorization": "Bearer access-token"}
    rv = test_client.post("/oauth/userinfo", headers=headers)
    assert rv.headers["Content-Type"] == "application/json"

    resp = json.loads(rv.data)
    assert resp == {
        "sub": "1",
        "address": {
            "country": "USA",
            "formatted": "742 Evergreen Terrace, Springfield",
            "locality": "Springfield",
            "postal_code": "1245",
            "region": "Unknown",
            "street_address": "742 Evergreen Terrace",
        },
        "birthdate": "2000-12-01",
        "email": "janedoe@example.com",
        "email_verified": True,
        "family_name": "Doe",
        "gender": "female",
        "given_name": "Jane",
        "locale": "fr-FR",
        "middle_name": "Middle",
        "name": "foo",
        "nickname": "Jany",
        "phone_number": "+1 (425) 555-1212",
        "phone_number_verified": False,
        "picture": "https://resource.test/janedoe/me.jpg",
        "preferred_username": "j.doe",
        "profile": "https://resource.test/janedoe",
        "updated_at": 1745315119,
        "website": "https://resource.test",
        "zoneinfo": "Europe/Paris",
    }


def test_no_token(test_client):
    rv = test_client.post("/oauth/userinfo")
    resp = json.loads(rv.data)
    assert resp["error"] == "missing_authorization"


def test_bad_token(test_client):
    headers = {"Authorization": "invalid token_string"}
    rv = test_client.post("/oauth/userinfo", headers=headers)
    resp = json.loads(rv.data)
    assert resp["error"] == "unsupported_token_type"


def test_token_has_bad_scope(test_client, db, token):
    """Test that tokens without 'openid' scope cannot access the userinfo endpoint."""

    token.scope = "foobar"
    db.session.add(token)
    db.session.commit()

    headers = {"Authorization": "Bearer access-token"}
    rv = test_client.post("/oauth/userinfo", headers=headers)
    resp = json.loads(rv.data)
    assert resp["error"] == "insufficient_scope"


def test_scope_minimum(test_client):
    headers = {"Authorization": "Bearer access-token"}
    rv = test_client.get("/oauth/userinfo", headers=headers)
    resp = json.loads(rv.data)
    assert resp == {
        "sub": "1",
    }


def test_scope_profile(test_client, db, token):
    token.scope = "openid profile"
    db.session.add(token)
    db.session.commit()

    headers = {"Authorization": "Bearer access-token"}
    rv = test_client.get("/oauth/userinfo", headers=headers)
    resp = json.loads(rv.data)
    assert resp == {
        "sub": "1",
        "birthdate": "2000-12-01",
        "family_name": "Doe",
        "gender": "female",
        "given_name": "Jane",
        "locale": "fr-FR",
        "middle_name": "Middle",
        "name": "foo",
        "nickname": "Jany",
        "picture": "https://resource.test/janedoe/me.jpg",
        "preferred_username": "j.doe",
        "profile": "https://resource.test/janedoe",
        "updated_at": 1745315119,
        "website": "https://resource.test",
        "zoneinfo": "Europe/Paris",
    }


def test_scope_address(test_client, db, token):
    token.scope = "openid address"
    db.session.add(token)
    db.session.commit()

    headers = {"Authorization": "Bearer access-token"}
    rv = test_client.get("/oauth/userinfo", headers=headers)
    resp = json.loads(rv.data)
    assert resp == {
        "sub": "1",
        "address": {
            "country": "USA",
            "formatted": "742 Evergreen Terrace, Springfield",
            "locality": "Springfield",
            "postal_code": "1245",
            "region": "Unknown",
            "street_address": "742 Evergreen Terrace",
        },
    }


def test_scope_email(test_client, db, token):
    token.scope = "openid email"
    db.session.add(token)
    db.session.commit()

    headers = {"Authorization": "Bearer access-token"}
    rv = test_client.get("/oauth/userinfo", headers=headers)
    resp = json.loads(rv.data)
    assert resp == {
        "sub": "1",
        "email": "janedoe@example.com",
        "email_verified": True,
    }


def test_scope_phone(test_client, db, token):
    token.scope = "openid phone"
    db.session.add(token)
    db.session.commit()

    headers = {"Authorization": "Bearer access-token"}
    rv = test_client.get("/oauth/userinfo", headers=headers)
    resp = json.loads(rv.data)
    assert resp == {
        "sub": "1",
        "phone_number": "+1 (425) 555-1212",
        "phone_number_verified": False,
    }


def test_scope_signed_unsecured(test_client, db, token, client):
    """When userinfo_signed_response_alg is set as client metadata, the userinfo response must be a JWT."""
    client.set_client_metadata(
        {
            "scope": "profile",
            "redirect_uris": ["https://client.test/authorized"],
            "userinfo_signed_response_alg": "none",
        }
    )
    db.session.add(client)
    db.session.commit()

    token.scope = "openid email"
    db.session.add(token)
    db.session.commit()

    headers = {"Authorization": "Bearer access-token"}
    rv = test_client.get("/oauth/userinfo", headers=headers)
    assert rv.headers["Content-Type"] == "application/jwt"

    jwt_none = JsonWebToken(["none"])
    claims = jwt_none.decode(rv.data, None)
    assert claims == {
        "sub": "1",
        "iss": "https://provider.test",
        "aud": "client-id",
        "email": "janedoe@example.com",
        "email_verified": True,
    }


def test_scope_signed_secured(test_client, client, token, db):
    """When userinfo_signed_response_alg is set as client metadata and not none, the userinfo response must be signed."""
    client.set_client_metadata(
        {
            "scope": "profile",
            "redirect_uris": ["https://client.test/authorized"],
            "userinfo_signed_response_alg": "RS256",
        }
    )
    db.session.add(client)
    db.session.commit()

    token.scope = "openid email"
    db.session.add(token)
    db.session.commit()

    headers = {"Authorization": "Bearer access-token"}
    rv = test_client.get("/oauth/userinfo", headers=headers)
    assert rv.headers["Content-Type"] == "application/jwt"

    pub_key = read_file_path("jwks_public.json")
    claims = jwt.decode(rv.data, pub_key)
    assert claims == {
        "sub": "1",
        "iss": "https://provider.test",
        "aud": "client-id",
        "email": "janedoe@example.com",
        "email_verified": True,
    }



import warnings


class AuthlibDeprecationWarning(DeprecationWarning):
    pass


warnings.simplefilter("always", AuthlibDeprecationWarning)


def deprecate(message, version=None, link_uid=None, link_file=None, stacklevel=3):
    if version:
        message += f"\nIt will be compatible before version {version}."

    if link_uid and link_file:
        message += f"\nRead more <https://git.io/{link_uid}#file-{link_file}-md>"

    warnings.warn(AuthlibDeprecationWarning(message), stacklevel=stacklevel)



import pytest
from flask import json

from authlib.oauth2.rfc6749.grants import ClientCredentialsGrant

from .oauth2_server import create_basic_header


@pytest.fixture(autouse=True)
def server(server):
    server.register_grant(ClientCredentialsGrant)
    return server


@pytest.fixture(autouse=True)
def client(client, db):
    client.set_client_metadata(
        {
            "scope": "profile",
            "redirect_uris": ["https://client.test/authorized"],
            "grant_types": ["client_credentials"],
        }
    )
    db.session.add(client)
    db.session.commit()
    return client


def test_invalid_client(test_client):
    rv = test_client.post(
        "/oauth/token",
        data={
            "grant_type": "client_credentials",
        },
    )
    resp = json.loads(rv.data)
    assert resp["error"] == "invalid_client"

    headers = create_basic_header("client-id", "invalid-secret")
    rv = test_client.post(
        "/oauth/token",
        data={
            "grant_type": "client_credentials",
        },
        headers=headers,
    )
    resp = json.loads(rv.data)
    assert resp["error"] == "invalid_client"


def test_invalid_grant_type(test_client, client, db):
    client.set_client_metadata(
        {
            "scope": "profile",
            "redirect_uris": ["https://client.test/authorized"],
            "grant_types": ["invalid"],
        }
    )
    db.session.add(client)
    db.session.commit()
    headers = create_basic_header("client-id", "client-secret")
    rv = test_client.post(
        "/oauth/token",
        data={
            "grant_type": "client_credentials",
        },
        headers=headers,
    )
    resp = json.loads(rv.data)
    assert resp["error"] == "unauthorized_client"


def test_invalid_scope(test_client, server):
    server.scopes_supported = ["profile"]
    headers = create_basic_header("client-id", "client-secret")
    rv = test_client.post(
        "/oauth/token",
        data={
            "grant_type": "client_credentials",
            "scope": "invalid",
        },
        headers=headers,
    )
    resp = json.loads(rv.data)
    assert resp["error"] == "invalid_scope"


def test_authorize_token(test_client):
    headers = create_basic_header("client-id", "client-secret")
    rv = test_client.post(
        "/oauth/token",
        data={
            "grant_type": "client_credentials",
        },
        headers=headers,
    )
    resp = json.loads(rv.data)
    assert "access_token" in resp


def test_token_generator(test_client, app, server):
    m = "tests.flask.test_oauth2.oauth2_server:token_generator"
    app.config.update({"OAUTH2_ACCESS_TOKEN_GENERATOR": m})
    server.load_config(app.config)

    headers = create_basic_header("client-id", "client-secret")
    rv = test_client.post(
        "/oauth/token",
        data={
            "grant_type": "client_credentials",
        },
        headers=headers,
    )
    resp = json.loads(rv.data)
    assert "access_token" in resp
    assert "c-client_credentials." in resp["access_token"]



import logging

from ..errors import UnauthorizedClientError
from ..hooks import hooked
from .base import BaseGrant
from .base import TokenEndpointMixin

log = logging.getLogger(__name__)


class ClientCredentialsGrant(BaseGrant, TokenEndpointMixin):
    """The client can request an access token using only its client
    credentials (or other supported means of authentication) when the
    client is requesting access to the protected resources under its
    control, or those of another resource owner that have been previously
    arranged with the authorization server.

    The client credentials grant type MUST only be used by confidential
    clients::

        +---------+                                  +---------------+
        |         |                                  |               |
        |         |>--(A)- Client Authentication --->| Authorization |
        | Client  |                                  |     Server    |
        |         |<--(B)---- Access Token ---------<|               |
        |         |                                  |               |
        +---------+                                  +---------------+

    https://tools.ietf.org/html/rfc6749#section-4.4
    """

    GRANT_TYPE = "client_credentials"

    def validate_token_request(self):
        """The client makes a request to the token endpoint by adding the
        following parameters using the "application/x-www-form-urlencoded"
        format per Appendix B with a character encoding of UTF-8 in the HTTP
        request entity-body:

        grant_type
             REQUIRED.  Value MUST be set to "client_credentials".

        scope
             OPTIONAL.  The scope of the access request as described by
             Section 3.3.

        The client MUST authenticate with the authorization server as
        described in Section 3.2.1.

        For example, the client makes the following HTTP request using
        transport-layer security (with extra line breaks for display purposes
        only):

        .. code-block:: http

            POST /token HTTP/1.1
            Host: server.example.com
            Authorization: Basic czZCaGRSa3F0MzpnWDFmQmF0M2JW
            Content-Type: application/x-www-form-urlencoded

            grant_type=client_credentials

        The authorization server MUST authenticate the client.
        """
        # ignore validate for grant_type, since it is validated by
        # check_token_endpoint
        client = self.authenticate_token_endpoint_client()
        log.debug("Validate token request of %r", client)

        if not client.check_grant_type(self.GRANT_TYPE):
            raise UnauthorizedClientError(
                f"The client is not authorized to use 'grant_type={self.GRANT_TYPE}'"
            )

        self.request.client = client
        self.validate_requested_scope()

    @hooked
    def create_token_response(self):
        """If the access token request is valid and authorized, the
        authorization server issues an access token as described in
        Section 5.1.  A refresh token SHOULD NOT be included.  If the request
        failed client authentication or is invalid, the authorization server
        returns an error response as described in Section 5.2.

        An example successful response:

        .. code-block:: http

            HTTP/1.1 200 OK
            Content-Type: application/json
            Cache-Control: no-store
            Pragma: no-cache

            {
                "access_token":"2YotnFZFEjr1zCsicMWpAA",
                "token_type":"example",
                "expires_in":3600,
                "example_parameter":"example_value"
            }

        :returns: (status_code, body, headers)
        """
        token = self.generate_token(
            scope=self.request.payload.scope, include_refresh_token=False
        )
        log.debug("Issue token %r to %r", token, self.client)
        self.save_token(token)
        return 200, token, self.TOKEN_RESPONSE_HEADER



"""authlib.oidc.discover.
~~~~~~~~~~~~~~~~~~~~~

OpenID Connect Discovery 1.0 Implementation.

https://openid.net/specs/openid-connect-discovery-1_0.html
"""

from .models import OpenIDProviderMetadata
from .well_known import get_well_known_url

__all__ = ["OpenIDProviderMetadata", "get_well_known_url"]



import os

import pytest

from authlib.integrations.django_oauth1 import CacheAuthorizationServer

from .models import Client
from .models import TokenCredential
from .models import User

pytestmark = pytest.mark.django_db


@pytest.fixture(autouse=True)
def env():
    os.environ["AUTHLIB_INSECURE_TRANSPORT"] = "true"
    yield
    del os.environ["AUTHLIB_INSECURE_TRANSPORT"]


@pytest.fixture
def server(settings):
    """Create server that respects current settings."""
    return CacheAuthorizationServer(Client, TokenCredential)


@pytest.fixture
def plaintext_server(settings):
    """Server configured with PLAINTEXT signature method."""
    settings.AUTHLIB_OAUTH1_PROVIDER = {"signature_methods": ["PLAINTEXT"]}
    return CacheAuthorizationServer(Client, TokenCredential)


@pytest.fixture
def rsa_server(settings):
    """Server configured with RSA-SHA1 signature method."""
    settings.AUTHLIB_OAUTH1_PROVIDER = {"signature_methods": ["RSA-SHA1"]}
    return CacheAuthorizationServer(Client, TokenCredential)


@pytest.fixture(autouse=True)
def user(db):
    user = User(username="foo")
    user.save()
    yield user
    user.delete()


@pytest.fixture(autouse=True)
def client(user, db):
    client = Client(
        user_id=user.pk,
        client_id="client",
        client_secret="secret",
        default_redirect_uri="https://client.test",
    )
    client.save()
    yield client
    client.delete()



from authlib.common.encoding import json_b64encode
from authlib.common.encoding import to_bytes
from authlib.common.encoding import to_unicode
from authlib.common.encoding import urlsafe_b64encode
from authlib.jose.errors import BadSignatureError
from authlib.jose.errors import DecodeError
from authlib.jose.errors import InvalidCritHeaderParameterNameError
from authlib.jose.errors import InvalidHeaderParameterNameError
from authlib.jose.errors import MissingAlgorithmError
from authlib.jose.errors import UnsupportedAlgorithmError
from authlib.jose.util import ensure_dict
from authlib.jose.util import extract_header
from authlib.jose.util import extract_segment

from .models import JWSHeader
from .models import JWSObject


class JsonWebSignature:
    #: Registered Header Parameter Names defined by Section 4.1
    REGISTERED_HEADER_PARAMETER_NAMES = frozenset(
        [
            "alg",
            "jku",
            "jwk",
            "kid",
            "x5u",
            "x5c",
            "x5t",
            "x5t#S256",
            "typ",
            "cty",
            "crit",
        ]
    )

    MAX_CONTENT_LENGTH: int = 256000

    #: Defined available JWS algorithms in the registry
    ALGORITHMS_REGISTRY = {}

    def __init__(self, algorithms=None, private_headers=None):
        self._private_headers = private_headers
        self._algorithms = algorithms

    @classmethod
    def register_algorithm(cls, algorithm):
        if not algorithm or algorithm.algorithm_type != "JWS":
            raise ValueError(f"Invalid algorithm for JWS, {algorithm!r}")
        cls.ALGORITHMS_REGISTRY[algorithm.name] = algorithm

    def serialize_compact(self, protected, payload, key):
        """Generate a JWS Compact Serialization. The JWS Compact Serialization
        represents digitally signed or MACed content as a compact, URL-safe
        string, per `Section 7.1`_.

        .. code-block:: text

            BASE64URL(UTF8(JWS Protected Header)) || '.' ||
            BASE64URL(JWS Payload) || '.' ||
            BASE64URL(JWS Signature)

        :param protected: A dict of protected header
        :param payload: A bytes/string of payload
        :param key: Private key used to generate signature
        :return: byte
        """
        jws_header = JWSHeader(protected, None)
        self._validate_private_headers(protected)
        self._validate_crit_headers(protected)
        algorithm, key = self._prepare_algorithm_key(protected, payload, key)

        protected_segment = json_b64encode(jws_header.protected)
        payload_segment = urlsafe_b64encode(to_bytes(payload))

        # calculate signature
        signing_input = b".".join([protected_segment, payload_segment])
        signature = urlsafe_b64encode(algorithm.sign(signing_input, key))
        return b".".join([protected_segment, payload_segment, signature])

    def deserialize_compact(self, s, key, decode=None):
        """Exact JWS Compact Serialization, and validate with the given key.
        If key is not provided, the returned dict will contain the signature,
        and signing input values. Via `Section 7.1`_.

        :param s: text of JWS Compact Serialization
        :param key: key used to verify the signature
        :param decode: a function to decode payload data
        :return: JWSObject
        :raise: BadSignatureError

        .. _`Section 7.1`: https://tools.ietf.org/html/rfc7515#section-7.1
        """
        if len(s) > self.MAX_CONTENT_LENGTH:
            raise ValueError("Serialization is too long.")

        try:
            s = to_bytes(s)
            signing_input, signature_segment = s.rsplit(b".", 1)
            protected_segment, payload_segment = signing_input.split(b".", 1)
        except ValueError as exc:
            raise DecodeError("Not enough segments") from exc

        protected = _extract_header(protected_segment)
        self._validate_crit_headers(protected)
        jws_header = JWSHeader(protected, None)

        payload = _extract_payload(payload_segment)
        if decode:
            payload = decode(payload)

        signature = _extract_signature(signature_segment)
        rv = JWSObject(jws_header, payload, "compact")
        algorithm, key = self._prepare_algorithm_key(jws_header, payload, key)
        if algorithm.verify(signing_input, signature, key):
            return rv
        raise BadSignatureError(rv)

    def serialize_json(self, header_obj, payload, key):
        """Generate a JWS JSON Serialization. The JWS JSON Serialization
        represents digitally signed or MACed content as a JSON object,
        per `Section 7.2`_.

        :param header_obj: A dict/list of header
        :param payload: A string/dict of payload
        :param key: Private key used to generate signature
        :return: JWSObject

        Example ``header_obj`` of JWS JSON Serialization::

            {
                "protected: {"alg": "HS256"},
                "header": {"kid": "jose"}
            }

        Pass a dict to generate flattened JSON Serialization, pass a list of
        header dict to generate standard JSON Serialization.
        """
        payload_segment = json_b64encode(payload)

        def _sign(jws_header):
            self._validate_private_headers(jws_header)
            # RFC 7515 §4.1.11: 'crit' MUST be integrity-protected.
            # Reject if present in unprotected header, and validate only
            # against the protected header parameters.
            self._reject_unprotected_crit(jws_header.header)
            self._validate_crit_headers(jws_header.protected)
            _alg, _key = self._prepare_algorithm_key(jws_header, payload, key)

            protected_segment = json_b64encode(jws_header.protected)
            signing_input = b".".join([protected_segment, payload_segment])
            signature = urlsafe_b64encode(_alg.sign(signing_input, _key))

            rv = {
                "protected": to_unicode(protected_segment),
                "signature": to_unicode(signature),
            }
            if jws_header.header is not None:
                rv["header"] = jws_header.header
            return rv

        if isinstance(header_obj, dict):
            data = _sign(JWSHeader.from_dict(header_obj))
            data["payload"] = to_unicode(payload_segment)
            return data

        signatures = [_sign(JWSHeader.from_dict(h)) for h in header_obj]
        return {"payload": to_unicode(payload_segment), "signatures": signatures}

    def deserialize_json(self, obj, key, decode=None):
        """Exact JWS JSON Serialization, and validate with the given key.
        If key is not provided, it will return a dict without signature
        verification. Header will still be validated. Via `Section 7.2`_.

        :param obj: text of JWS JSON Serialization
        :param key: key used to verify the signature
        :param decode: a function to decode payload data
        :return: JWSObject
        :raise: BadSignatureError

        .. _`Section 7.2`: https://tools.ietf.org/html/rfc7515#section-7.2
        """
        obj = ensure_dict(obj, "JWS")

        payload_segment = obj.get("payload")
        if payload_segment is None:
            raise DecodeError('Missing "payload" value')

        payload_segment = to_bytes(payload_segment)
        payload = _extract_payload(payload_segment)
        if decode:
            payload = decode(payload)

        if "signatures" not in obj:
            # flattened JSON JWS
            jws_header, valid = self._validate_json_jws(
                payload_segment, payload, obj, key
            )

            rv = JWSObject(jws_header, payload, "flat")
            if valid:
                return rv
            raise BadSignatureError(rv)

        headers = []
        is_valid = True
        for header_obj in obj["signatures"]:
            jws_header, valid = self._validate_json_jws(
                payload_segment, payload, header_obj, key
            )
            headers.append(jws_header)
            if not valid:
                is_valid = False

        rv = JWSObject(headers, payload, "json")
        if is_valid:
            return rv
        raise BadSignatureError(rv)

    def serialize(self, header, payload, key):
        """Generate a JWS Serialization. It will automatically generate a
        Compact or JSON Serialization depending on the given header. If a
        header is in a JSON header format, it will call
        :meth:`serialize_json`, otherwise it will call
        :meth:`serialize_compact`.

        :param header: A dict/list of header
        :param payload: A string/dict of payload
        :param key: Private key used to generate signature
        :return: byte/dict
        """
        if isinstance(header, (list, tuple)):
            return self.serialize_json(header, payload, key)
        if "protected" in header:
            return self.serialize_json(header, payload, key)
        return self.serialize_compact(header, payload, key)

    def deserialize(self, s, key, decode=None):
        """Deserialize JWS Serialization, both compact and JSON format.
        It will automatically deserialize depending on the given JWS.

        :param s: text of JWS Compact/JSON Serialization
        :param key: key used to verify the signature
        :param decode: a function to decode payload data
        :return: dict
        :raise: BadSignatureError

        If key is not provided, it will still deserialize the serialization
        without verification.
        """
        if isinstance(s, dict):
            return self.deserialize_json(s, key, decode)

        s = to_bytes(s)
        if s.startswith(b"{") and s.endswith(b"}"):
            return self.deserialize_json(s, key, decode)
        return self.deserialize_compact(s, key, decode)

    def _prepare_algorithm_key(self, header, payload, key):
        if "alg" not in header:
            raise MissingAlgorithmError()

        alg = header["alg"]
        if self._algorithms is not None and alg not in self._algorithms:
            raise UnsupportedAlgorithmError()
        if alg not in self.ALGORITHMS_REGISTRY:
            raise UnsupportedAlgorithmError()

        algorithm = self.ALGORITHMS_REGISTRY[alg]
        if callable(key):
            key = key(header, payload)
        elif key is None and "jwk" in header:
            key = header["jwk"]
        key = algorithm.prepare_key(key)
        return algorithm, key

    def _validate_private_headers(self, header):
        # only validate private headers when developers set
        # private headers explicitly
        if self._private_headers is not None:
            names = self.REGISTERED_HEADER_PARAMETER_NAMES.copy()
            names = names.union(self._private_headers)

            for k in header:
                if k not in names:
                    raise InvalidHeaderParameterNameError(k)

    def _reject_unprotected_crit(self, unprotected_header):
        """Reject 'crit' when found in the unprotected header (RFC 7515 §4.1.11)."""
        if unprotected_header and "crit" in unprotected_header:
            raise InvalidHeaderParameterNameError("crit")

    def _validate_crit_headers(self, header):
        if "crit" in header:
            crit_headers = header["crit"]
            # Type enforcement for robustness and predictable errors
            if not isinstance(crit_headers, list) or not all(
                isinstance(x, str) for x in crit_headers
            ):
                raise InvalidHeaderParameterNameError("crit")
            names = self.REGISTERED_HEADER_PARAMETER_NAMES.copy()
            if self._private_headers:
                names = names.union(self._private_headers)
            for k in crit_headers:
                if k not in names:
                    raise InvalidCritHeaderParameterNameError(k)
                elif k not in header:
                    raise InvalidCritHeaderParameterNameError(k)

    def _validate_json_jws(self, payload_segment, payload, header_obj, key):
        protected_segment = header_obj.get("protected")
        if not protected_segment:
            raise DecodeError('Missing "protected" value')

        signature_segment = header_obj.get("signature")
        if not signature_segment:
            raise DecodeError('Missing "signature" value')

        protected_segment = to_bytes(protected_segment)
        protected = _extract_header(protected_segment)
        header = header_obj.get("header")
        if header and not isinstance(header, dict):
            raise DecodeError('Invalid "header" value')
        # RFC 7515 §4.1.11: 'crit' MUST be integrity-protected. If present in
        # the unprotected header object, reject the JWS.
        self._reject_unprotected_crit(header)

        # Enforce must-understand semantics for names listed in protected
        # 'crit'. This will also ensure each listed name is present in the
        # protected header.
        self._validate_crit_headers(protected)
        jws_header = JWSHeader(protected, header)
        algorithm, key = self._prepare_algorithm_key(jws_header, payload, key)
        signing_input = b".".join([protected_segment, payload_segment])
        signature = _extract_signature(to_bytes(signature_segment))
        if algorithm.verify(signing_input, signature, key):
            return jws_header, True
        return jws_header, False


def _extract_header(header_segment):
    return extract_header(header_segment, DecodeError)


def _extract_signature(signature_segment):
    return extract_segment(signature_segment, DecodeError, "signature")


def _extract_payload(payload_segment):
    return extract_segment(payload_segment, DecodeError, "payload")



import time

from authlib.common.encoding import to_native
from authlib.common.urls import add_params_to_uri
from authlib.common.urls import quote_url
from authlib.jose import JsonWebToken
from authlib.oauth2.rfc6749 import InvalidRequestError
from authlib.oauth2.rfc6749 import scope_to_list

from ..errors import AccountSelectionRequiredError
from ..errors import ConsentRequiredError
from ..errors import LoginRequiredError
from ..util import create_half_hash


def is_openid_scope(scope):
    scopes = scope_to_list(scope)
    return scopes and "openid" in scopes


def validate_request_prompt(grant, redirect_uri, redirect_fragment=False):
    prompt = grant.request.payload.data.get("prompt")
    end_user = grant.request.user
    if not prompt:
        if not end_user:
            grant.prompt = "login"
        return grant

    if prompt == "none" and not end_user:
        raise LoginRequiredError(
            redirect_uri=redirect_uri, redirect_fragment=redirect_fragment
        )

    prompts = prompt.split()
    if "none" in prompts and len(prompts) > 1:
        # If this parameter contains none with any other value,
        # an error is returned
        raise InvalidRequestError(
            "Invalid 'prompt' parameter.",
            redirect_uri=redirect_uri,
            redirect_fragment=redirect_fragment,
        )

    prompt = _guess_prompt_value(
        end_user, prompts, redirect_uri, redirect_fragment=redirect_fragment
    )
    if prompt:
        grant.prompt = prompt
    return grant


def validate_nonce(request, exists_nonce, required=False):
    nonce = request.payload.data.get("nonce")
    if not nonce:
        if required:
            raise InvalidRequestError("Missing 'nonce' in request.")
        return True

    if exists_nonce(nonce, request):
        raise InvalidRequestError("Replay attack")


def generate_id_token(
    token,
    user_info,
    key,
    iss,
    aud,
    alg="RS256",
    exp=3600,
    nonce=None,
    auth_time=None,
    acr=None,
    amr=None,
    code=None,
    kid=None,
):
    now = int(time.time())
    if auth_time is None:
        auth_time = now

    header = {"alg": alg}
    if kid:
        header["kid"] = kid

    payload = {
        "iss": iss,
        "aud": aud,
        "iat": now,
        "exp": now + exp,
        "auth_time": auth_time,
    }
    if nonce:
        payload["nonce"] = nonce

    if acr:
        payload["acr"] = acr

    if amr:
        payload["amr"] = amr

    if code:
        c_hash = create_half_hash(code, alg)
        if c_hash is not None:
            payload["c_hash"] = to_native(c_hash)

    access_token = token.get("access_token")
    if access_token:
        at_hash = create_half_hash(access_token, alg)
        if at_hash is not None:
            payload["at_hash"] = to_native(at_hash)

    payload.update(user_info)
    return to_native(JsonWebToken([alg]).encode(header, payload, key))


def create_response_mode_response(redirect_uri, params, response_mode):
    if response_mode == "form_post":
        tpl = (
            "<html><head><title>Redirecting</title></head>"
            '<body onload="javascript:document.forms[0].submit()">'
            '<form method="post" action="{}">{}</form></body></html>'
        )
        inputs = "".join(
            [
                f'<input type="hidden" name="{quote_url(k)}" value="{quote_url(v)}"/>'
                for k, v in params
            ]
        )
        body = tpl.format(quote_url(redirect_uri), inputs)
        return 200, body, [("Content-Type", "text/html; charset=utf-8")]

    if response_mode == "query":
        uri = add_params_to_uri(redirect_uri, params, fragment=False)
    elif response_mode == "fragment":
        uri = add_params_to_uri(redirect_uri, params, fragment=True)
    else:
        raise InvalidRequestError('Invalid "response_mode" value')

    return 302, "", [("Location", uri)]


def _guess_prompt_value(end_user, prompts, redirect_uri, redirect_fragment):
    # http://openid.net/specs/openid-connect-core-1_0.html#AuthRequest

    if not end_user or "login" in prompts:
        return "login"

    if "consent" in prompts:
        if not end_user:
            raise ConsentRequiredError(
                redirect_uri=redirect_uri, redirect_fragment=redirect_fragment
            )
        return "consent"
    elif "select_account" in prompts:
        if not end_user:
            raise AccountSelectionRequiredError(
                redirect_uri=redirect_uri, redirect_fragment=redirect_fragment
            )
        return "select_account"



from authlib.jose import JsonWebKey
from authlib.jose import JsonWebToken
from authlib.oidc.core import CodeIDToken
from authlib.oidc.core import ImplicitIDToken
from authlib.oidc.core import UserInfo

__all__ = ["AsyncOpenIDMixin"]


class AsyncOpenIDMixin:
    async def fetch_jwk_set(self, force=False):
        metadata = await self.load_server_metadata()
        jwk_set = metadata.get("jwks")
        if jwk_set and not force:
            return jwk_set

        uri = metadata.get("jwks_uri")
        if not uri:
            raise RuntimeError('Missing "jwks_uri" in metadata')

        async with self.client_cls(**self.client_kwargs) as client:
            resp = await client.request("GET", uri, withhold_token=True)
            resp.raise_for_status()
            jwk_set = resp.json()

        self.server_metadata["jwks"] = jwk_set
        return jwk_set

    async def userinfo(self, **kwargs):
        """Fetch user info from ``userinfo_endpoint``."""
        metadata = await self.load_server_metadata()
        resp = await self.get(metadata["userinfo_endpoint"], **kwargs)
        resp.raise_for_status()
        data = resp.json()
        return UserInfo(data)

    async def parse_id_token(
        self, token, nonce, claims_options=None, claims_cls=None, leeway=120
    ):
        """Return an instance of UserInfo from token's ``id_token``."""
        claims_params = dict(
            nonce=nonce,
            client_id=self.client_id,
        )
        if claims_cls is None:
            if "access_token" in token:
                claims_params["access_token"] = token["access_token"]
                claims_cls = CodeIDToken
            else:
                claims_cls = ImplicitIDToken

        metadata = await self.load_server_metadata()
        if claims_options is None and "issuer" in metadata:
            claims_options = {"iss": {"values": [metadata["issuer"]]}}

        alg_values = metadata.get("id_token_signing_alg_values_supported")
        if not alg_values:
            alg_values = ["RS256"]

        jwt = JsonWebToken(alg_values)

        jwk_set = await self.fetch_jwk_set()
        try:
            claims = jwt.decode(
                token["id_token"],
                key=JsonWebKey.import_key_set(jwk_set),
                claims_cls=claims_cls,
                claims_options=claims_options,
                claims_params=claims_params,
            )
        except ValueError:
            jwk_set = await self.fetch_jwk_set(force=True)
            claims = jwt.decode(
                token["id_token"],
                key=JsonWebKey.import_key_set(jwk_set),
                claims_cls=claims_cls,
                claims_options=claims_options,
                claims_params=claims_params,
            )

        # https://github.com/authlib/authlib/issues/259
        if claims.get("nonce_supported") is False:
            claims.params["nonce"] = None
        claims.validate(leeway=leeway)
        return UserInfo(claims)



"""authlib.oidc.core.grants.code.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Implementation of Authentication using the Authorization Code Flow
per `Section 3.1`_.

.. _`Section 3.1`: http://openid.net/specs/openid-connect-core-1_0.html#CodeFlowAuth
"""

import logging
import warnings

from authlib.oauth2.rfc6749 import OAuth2Request

from .util import generate_id_token
from .util import is_openid_scope
from .util import validate_nonce
from .util import validate_request_prompt

log = logging.getLogger(__name__)


class OpenIDToken:
    def get_jwt_config(self, grant, client):  # pragma: no cover
        """Get the JWT configuration for OpenIDCode extension. The JWT
        configuration will be used to generate ``id_token``.
        If ``alg`` is undefined, the ``id_token_signed_response_alg`` client
        metadata will be used. By default ``RS256`` will be used.
        If ``key`` is undefined, the ``jwks_uri`` or ``jwks`` client metadata
        will be used.
        Developers MUST implement this method in subclass, e.g.::

            def get_jwt_config(self, grant, client):
                return {
                    "key": read_private_key_file(key_path),
                    "alg": client.id_token_signed_response_alg or "RS256",
                    "iss": "issuer-identity",
                    "exp": 3600,
                }

        :param grant: AuthorizationCodeGrant instance
        :param client: OAuth2 client instance
        :return: dict
        """
        raise NotImplementedError()

    def generate_user_info(self, user, scope):
        """Provide user information for the given scope. Developers
        MUST implement this method in subclass, e.g.::

            from authlib.oidc.core import UserInfo


            def generate_user_info(self, user, scope):
                user_info = UserInfo(sub=user.id, name=user.name)
                if "email" in scope:
                    user_info["email"] = user.email
                return user_info

        :param user: user instance
        :param scope: scope of the token
        :return: ``authlib.oidc.core.UserInfo`` instance
        """
        raise NotImplementedError()

    def get_audiences(self, request):
        """Parse `aud` value for id_token, default value is client id. Developers
        MAY rewrite this method to provide a customized audience value.
        """
        client = request.client
        return [client.get_client_id()]

    def process_token(self, grant, response):
        _, token, _ = response
        scope = token.get("scope")
        if not scope or not is_openid_scope(scope):
            # standard authorization code flow
            return token

        request: OAuth2Request = grant.request
        authorization_code = request.authorization_code

        try:
            config = self.get_jwt_config(grant, request.client)
        except TypeError:
            warnings.warn(
                "get_jwt_config(self, grant) is deprecated and will be removed in version 1.8. "
                "Use get_jwt_config(self, grant, client) instead.",
                DeprecationWarning,
                stacklevel=2,
            )
            config = self.get_jwt_config(grant)

        config["aud"] = self.get_audiences(request)

        # Per OpenID Connect Registration 1.0 Section 2:
        # Use client's id_token_signed_response_alg if specified
        if not config.get("alg") and (
            client_alg := request.client.id_token_signed_response_alg
        ):
            config["alg"] = client_alg

        if authorization_code:
            config["nonce"] = authorization_code.get_nonce()
            config["auth_time"] = authorization_code.get_auth_time()

            if acr := authorization_code.get_acr():
                config["acr"] = acr

            if amr := authorization_code.get_amr():
                config["amr"] = amr

        user_info = self.generate_user_info(request.user, token["scope"])
        id_token = generate_id_token(token, user_info, **config)
        token["id_token"] = id_token
        return token

    def __call__(self, grant):
        grant.register_hook("after_create_token_response", self.process_token)


class OpenIDCode(OpenIDToken):
    """An extension from OpenID Connect for "grant_type=code" request. Developers
    MUST implement the missing methods::

        class MyOpenIDCode(OpenIDCode):
            def get_jwt_config(self, grant):
                return {...}

            def exists_nonce(self, nonce, request):
                return check_if_nonce_in_cache(request.payload.client_id, nonce)

            def generate_user_info(self, user, scope):
                return {...}

    The register this extension with AuthorizationCodeGrant::

        authorization_server.register_grant(
            AuthorizationCodeGrant, extensions=[MyOpenIDCode()]
        )
    """

    def __init__(self, require_nonce=False):
        self.require_nonce = require_nonce

    def exists_nonce(self, nonce, request):
        """Check if the given nonce is existing in your database. Developers
        MUST implement this method in subclass, e.g.::

            def exists_nonce(self, nonce, request):
                exists = AuthorizationCode.query.filter_by(
                    client_id=request.payload.client_id, nonce=nonce
                ).first()
                return bool(exists)

        :param nonce: A string of "nonce" parameter in request
        :param request: OAuth2Request instance
        :return: Boolean
        """
        raise NotImplementedError()

    def validate_openid_authorization_request(self, grant, redirect_uri):
        validate_nonce(grant.request, self.exists_nonce, self.require_nonce)

    def __call__(self, grant):
        grant.register_hook("after_create_token_response", self.process_token)
        if is_openid_scope(grant.request.payload.scope):
            grant.register_hook(
                "after_validate_authorization_request_payload",
                self.validate_openid_authorization_request,
            )
            grant.register_hook(
                "after_validate_consent_request", validate_request_prompt
            )



from authlib.common.errors import AuthlibHTTPError
from authlib.common.urls import add_params_to_uri


def invalid_error_characters(text: str) -> list[str]:
    """Check whether the string only contains characters from the restricted ASCII set defined in RFC6749 for errors.

    https://datatracker.ietf.org/doc/html/rfc6749#section-4.1.2.1
    """
    valid_ranges = [
        (0x20, 0x21),
        (0x23, 0x5B),
        (0x5D, 0x7E),
    ]

    return [
        char
        for char in set(text)
        if not any(start <= ord(char) <= end for start, end in valid_ranges)
    ]


class OAuth2Error(AuthlibHTTPError):
    def __init__(
        self,
        description=None,
        uri=None,
        status_code=None,
        state=None,
        redirect_uri=None,
        redirect_fragment=False,
        error=None,
    ):
        # Human-readable ASCII [USASCII] text providing
        # additional information, used to assist the client developer in
        # understanding the error that occurred.
        # Values for the "error_description" parameter MUST NOT include
        # characters outside the set %x20-21 / %x23-5B / %x5D-7E.
        if description:
            if chars := invalid_error_characters(description):
                raise ValueError(
                    f"Error description contains forbidden characters: {', '.join(chars)}."
                )

        super().__init__(error, description, uri, status_code)
        self.state = state
        self.redirect_uri = redirect_uri
        self.redirect_fragment = redirect_fragment

    def get_body(self):
        """Get a list of body."""
        error = super().get_body()
        if self.state:
            error.append(("state", self.state))
        return error

    def __call__(self, uri=None):
        if self.redirect_uri:
            params = self.get_body()
            loc = add_params_to_uri(self.redirect_uri, params, self.redirect_fragment)
            return 302, "", [("Location", loc)]
        return super().__call__(uri=uri)



import pytest

from authlib.oauth2.rfc6749.grants import (
    AuthorizationCodeGrant as _AuthorizationCodeGrant,
)
from authlib.oauth2.rfc9207 import IssuerParameter as _IssuerParameter

from .models import CodeGrantMixin
from .models import save_authorization_code

authorize_url = "/oauth/authorize?response_type=code&client_id=client-id"


class AuthorizationCodeGrant(CodeGrantMixin, _AuthorizationCodeGrant):
    TOKEN_ENDPOINT_AUTH_METHODS = ["client_secret_basic", "client_secret_post", "none"]

    def save_authorization_code(self, code, request):
        return save_authorization_code(code, request)


class IssuerParameter(_IssuerParameter):
    def get_issuer(self) -> str:
        return "https://auth.test"


@pytest.fixture(autouse=True)
def server(server):
    server.register_grant(AuthorizationCodeGrant)
    return server


@pytest.fixture(autouse=True)
def client(client, db):
    client.set_client_metadata(
        {
            "redirect_uris": ["https://client.test"],
            "scope": "profile address",
            "token_endpoint_auth_method": "client_secret_basic",
            "response_types": ["code"],
            "grant_types": ["authorization_code"],
        }
    )
    db.session.add(client)
    db.session.commit()
    return client


def test_rfc9207_enabled_success(test_client, server):
    """Check that when RFC9207 is implemented,
    the authorization response has an ``iss`` parameter."""

    server.register_extension(IssuerParameter())
    url = authorize_url + "&state=bar"
    rv = test_client.post(url, data={"user_id": "1"})
    assert "iss=https%3A%2F%2Fauth.test" in rv.location


def test_rfc9207_disabled_success_no_iss(test_client):
    """Check that when RFC9207 is not implemented,
    the authorization response contains no ``iss`` parameter."""

    url = authorize_url + "&state=bar"
    rv = test_client.post(url, data={"user_id": "1"})
    assert "iss=" not in rv.location


def test_rfc9207_enabled_error(test_client, server):
    """Check that when RFC9207 is implemented,
    the authorization response has an ``iss`` parameter,
    even when an error is returned."""

    server.register_extension(IssuerParameter())
    rv = test_client.post(authorize_url)
    assert "error=access_denied" in rv.location
    assert "iss=https%3A%2F%2Fauth.test" in rv.location


def test_rfc9207_disbled_error_no_iss(test_client):
    """Check that when RFC9207 is not implemented,
    the authorization response contains no ``iss`` parameter,
    even when an error is returned."""

    rv = test_client.post(authorize_url)
    assert "error=access_denied" in rv.location
    assert "iss=" not in rv.location



import json

import pytest

from authlib.integrations.django_oauth2 import RevocationEndpoint

from .models import Client
from .oauth2_server import create_basic_auth

ENDPOINT_NAME = RevocationEndpoint.ENDPOINT_NAME


@pytest.fixture(autouse=True)
def server(server):
    server.register_endpoint(RevocationEndpoint)
    return server


@pytest.fixture(autouse=True)
def client(user):
    client = Client(
        user_id=user.pk,
        client_id="client-id",
        client_secret="client-secret",
        token_endpoint_auth_method="client_secret_basic",
        default_redirect_uri="https://client.test",
    )
    client.save()
    yield client
    client.delete()


def test_invalid_client(factory, server):
    request = factory.post("/oauth/revoke")
    resp = server.create_endpoint_response(ENDPOINT_NAME, request)
    data = json.loads(resp.content)
    assert data["error"] == "invalid_client"

    request = factory.post("/oauth/revoke", HTTP_AUTHORIZATION="invalid token")
    resp = server.create_endpoint_response(ENDPOINT_NAME, request)
    data = json.loads(resp.content)
    assert data["error"] == "invalid_client"

    request = factory.post(
        "/oauth/revoke",
        HTTP_AUTHORIZATION=create_basic_auth("invalid", "client-secret"),
    )
    resp = server.create_endpoint_response(ENDPOINT_NAME, request)
    data = json.loads(resp.content)
    assert data["error"] == "invalid_client"

    request = factory.post(
        "/oauth/revoke",
        HTTP_AUTHORIZATION=create_basic_auth("client-id", "invalid"),
    )
    resp = server.create_endpoint_response(ENDPOINT_NAME, request)
    data = json.loads(resp.content)
    assert data["error"] == "invalid_client"


def test_invalid_token(factory, server, token):
    auth_header = create_basic_auth("client-id", "client-secret")

    request = factory.post("/oauth/revoke", HTTP_AUTHORIZATION=auth_header)
    resp = server.create_endpoint_response(ENDPOINT_NAME, request)
    data = json.loads(resp.content)
    assert data["error"] == "invalid_request"

    # case 1
    request = factory.post(
        "/oauth/revoke",
        data={"token": "invalid-token"},
        HTTP_AUTHORIZATION=auth_header,
    )
    resp = server.create_endpoint_response(ENDPOINT_NAME, request)
    assert resp.status_code == 200

    # case 2
    request = factory.post(
        "/oauth/revoke",
        data={
            "token": "a1",
            "token_type_hint": "unsupported_token_type",
        },
        HTTP_AUTHORIZATION=auth_header,
    )
    resp = server.create_endpoint_response(ENDPOINT_NAME, request)
    data = json.loads(resp.content)
    assert data["error"] == "unsupported_token_type"

    # case 3
    request = factory.post(
        "/oauth/revoke",
        data={
            "token": "a1",
            "token_type_hint": "refresh_token",
        },
        HTTP_AUTHORIZATION=auth_header,
    )
    resp = server.create_endpoint_response(ENDPOINT_NAME, request)
    assert resp.status_code == 200


def test_revoke_token_with_hint(factory, server, token):
    revoke_token(server, factory, {"token": "a1", "token_type_hint": "access_token"})
    revoke_token(server, factory, {"token": "r1", "token_type_hint": "refresh_token"})


def test_revoke_token_without_hint(factory, server, token):
    revoke_token(server, factory, {"token": "a1"})
    revoke_token(server, factory, {"token": "r1"})


def revoke_token(server, factory, data):
    auth_header = create_basic_auth("client-id", "client-secret")

    request = factory.post(
        "/oauth/revoke",
        data=data,
        HTTP_AUTHORIZATION=auth_header,
    )
    resp = server.create_endpoint_response(ENDPOINT_NAME, request)
    assert resp.status_code == 200



"""authlib.jose.rfc7519.
~~~~~~~~~~~~~~~~~~~~

This module represents a direct implementation of
JSON Web Token (JWT).

https://tools.ietf.org/html/rfc7519
"""

from .claims import BaseClaims
from .claims import JWTClaims
from .jwt import JsonWebToken

__all__ = ["JsonWebToken", "BaseClaims", "JWTClaims"]



import base64

from authlib.common.encoding import to_bytes
from authlib.common.encoding import to_native
from authlib.common.urls import add_params_to_qs
from authlib.common.urls import add_params_to_uri

from .rfc6749 import OAuth2Token
from .rfc6750 import add_bearer_token


def encode_client_secret_basic(client, method, uri, headers, body):
    text = f"{client.client_id}:{client.client_secret}"
    auth = to_native(base64.b64encode(to_bytes(text, "latin1")))
    headers["Authorization"] = f"Basic {auth}"
    return uri, headers, body


def encode_client_secret_post(client, method, uri, headers, body):
    body = add_params_to_qs(
        body or "",
        [
            ("client_id", client.client_id),
            ("client_secret", client.client_secret or ""),
        ],
    )
    if "Content-Length" in headers:
        headers["Content-Length"] = str(len(body))
    return uri, headers, body


def encode_none(client, method, uri, headers, body):
    if method == "GET":
        uri = add_params_to_uri(uri, [("client_id", client.client_id)])
        return uri, headers, body
    body = add_params_to_qs(body, [("client_id", client.client_id)])
    if "Content-Length" in headers:
        headers["Content-Length"] = str(len(body))
    return uri, headers, body


class ClientAuth:
    """Attaches OAuth Client Information to HTTP requests.

    :param client_id: Client ID, which you get from client registration.
    :param client_secret: Client Secret, which you get from registration.
    :param auth_method: Client auth method for token endpoint. The supported
        methods for now:

        * client_secret_basic (default)
        * client_secret_post
        * none
    """

    DEFAULT_AUTH_METHODS = {
        "client_secret_basic": encode_client_secret_basic,
        "client_secret_post": encode_client_secret_post,
        "none": encode_none,
    }

    def __init__(self, client_id, client_secret, auth_method=None):
        if auth_method is None:
            auth_method = "client_secret_basic"

        self.client_id = client_id
        self.client_secret = client_secret

        if auth_method in self.DEFAULT_AUTH_METHODS:
            auth_method = self.DEFAULT_AUTH_METHODS[auth_method]

        self.auth_method = auth_method

    def prepare(self, method, uri, headers, body):
        return self.auth_method(self, method, uri, headers, body)


class TokenAuth:
    """Attach token information to HTTP requests.

    :param token: A dict or OAuth2Token instance of an OAuth 2.0 token
    :param token_placement: The placement of the token, default is ``header``,
        available choices:

        * header (default)
        * body
        * uri
    """

    DEFAULT_TOKEN_TYPE = "bearer"
    SIGN_METHODS = {"bearer": add_bearer_token}

    def __init__(self, token, token_placement="header", client=None):
        self.token = OAuth2Token.from_dict(token)
        self.token_placement = token_placement
        self.client = client
        self.hooks = set()

    def set_token(self, token):
        self.token = OAuth2Token.from_dict(token)

    def prepare(self, uri, headers, body):
        token_type = self.token.get("token_type", self.DEFAULT_TOKEN_TYPE)
        sign = self.SIGN_METHODS[token_type.lower()]
        uri, headers, body = sign(
            self.token["access_token"], uri, headers, body, self.token_placement
        )

        for hook in self.hooks:
            uri, headers, body = hook(uri, headers, body)

        return uri, headers, body

    def __del__(self):
        del self.client
        del self.hooks



import pytest
from httpx import WSGITransport

from authlib.integrations.httpx_client import SIGNATURE_TYPE_BODY
from authlib.integrations.httpx_client import SIGNATURE_TYPE_QUERY
from authlib.integrations.httpx_client import OAuth1Client
from authlib.integrations.httpx_client import OAuthError

from ..wsgi_helper import MockDispatch

oauth_url = "https://provider.test/oauth"


def test_fetch_request_token_via_header():
    request_token = {"oauth_token": "1", "oauth_token_secret": "2"}

    def assert_func(request):
        auth_header = request.headers.get("authorization")
        assert 'oauth_consumer_key="id"' in auth_header
        assert "oauth_signature=" in auth_header

    transport = WSGITransport(MockDispatch(request_token, assert_func=assert_func))
    with OAuth1Client("id", "secret", transport=transport) as client:
        response = client.fetch_request_token(oauth_url)

    assert response == request_token


def test_fetch_request_token_via_body():
    request_token = {"oauth_token": "1", "oauth_token_secret": "2"}

    def assert_func(request):
        auth_header = request.headers.get("authorization")
        assert auth_header is None

        content = request.form
        assert content.get("oauth_consumer_key") == "id"
        assert "oauth_signature" in content

    transport = WSGITransport(MockDispatch(request_token, assert_func=assert_func))

    with OAuth1Client(
        "id",
        "secret",
        signature_type=SIGNATURE_TYPE_BODY,
        transport=transport,
    ) as client:
        response = client.fetch_request_token(oauth_url)

    assert response == request_token


def test_fetch_request_token_via_query():
    request_token = {"oauth_token": "1", "oauth_token_secret": "2"}

    def assert_func(request):
        auth_header = request.headers.get("authorization")
        assert auth_header is None

        url = str(request.url)
        assert "oauth_consumer_key=id" in url
        assert "&oauth_signature=" in url

    transport = WSGITransport(MockDispatch(request_token, assert_func=assert_func))

    with OAuth1Client(
        "id",
        "secret",
        signature_type=SIGNATURE_TYPE_QUERY,
        transport=transport,
    ) as client:
        response = client.fetch_request_token(oauth_url)

    assert response == request_token


def test_fetch_access_token():
    request_token = {"oauth_token": "1", "oauth_token_secret": "2"}

    def assert_func(request):
        auth_header = request.headers.get("authorization")
        assert 'oauth_verifier="d"' in auth_header
        assert 'oauth_token="foo"' in auth_header
        assert 'oauth_consumer_key="id"' in auth_header
        assert "oauth_signature=" in auth_header

    transport = WSGITransport(MockDispatch(request_token, assert_func=assert_func))
    with OAuth1Client(
        "id",
        "secret",
        token="foo",
        token_secret="bar",
        transport=transport,
    ) as client:
        with pytest.raises(OAuthError):
            client.fetch_access_token(oauth_url)

        response = client.fetch_access_token(oauth_url, verifier="d")

    assert response == request_token


def test_get_via_header():
    transport = WSGITransport(MockDispatch(b"hello"))
    with OAuth1Client(
        "id",
        "secret",
        token="foo",
        token_secret="bar",
        transport=transport,
    ) as client:
        response = client.get("https://resource.test/")

    assert response.content == b"hello"
    request = response.request
    auth_header = request.headers.get("authorization")
    assert 'oauth_token="foo"' in auth_header
    assert 'oauth_consumer_key="id"' in auth_header
    assert "oauth_signature=" in auth_header


def test_get_via_body():
    def assert_func(request):
        content = request.form
        assert content.get("oauth_token") == "foo"
        assert content.get("oauth_consumer_key") == "id"
        assert "oauth_signature" in content

    transport = WSGITransport(MockDispatch(b"hello", assert_func=assert_func))
    with OAuth1Client(
        "id",
        "secret",
        token="foo",
        token_secret="bar",
        signature_type=SIGNATURE_TYPE_BODY,
        transport=transport,
    ) as client:
        response = client.post("https://resource.test/")

    assert response.content == b"hello"

    request = response.request
    auth_header = request.headers.get("authorization")
    assert auth_header is None


def test_get_via_query():
    transport = WSGITransport(MockDispatch(b"hello"))
    with OAuth1Client(
        "id",
        "secret",
        token="foo",
        token_secret="bar",
        signature_type=SIGNATURE_TYPE_QUERY,
        transport=transport,
    ) as client:
        response = client.get("https://resource.test/")

    assert response.content == b"hello"
    request = response.request
    auth_header = request.headers.get("authorization")
    assert auth_header is None

    url = str(request.url)
    assert "oauth_token=foo" in url
    assert "oauth_consumer_key=id" in url
    assert "oauth_signature=" in url



import base64

from flask import Flask
from flask import request

from authlib.common.encoding import to_bytes
from authlib.common.encoding import to_unicode
from authlib.common.security import generate_token
from authlib.integrations.flask_oauth2 import AuthorizationServer
from authlib.integrations.sqla_oauth2 import create_query_client_func
from authlib.integrations.sqla_oauth2 import create_save_token_func
from authlib.oauth2 import OAuth2Error

from .models import Client
from .models import Token
from .models import User
from .models import db


def token_generator(client, grant_type, user=None, scope=None):
    token = f"{client.client_id[0]}-{grant_type}"
    if user:
        token = f"{token}.{user.get_user_id()}"
    return f"{token}.{generate_token(32)}"


def create_authorization_server(app, lazy=False):
    query_client = create_query_client_func(db.session, Client)
    save_token = create_save_token_func(db.session, Token)

    if lazy:
        server = AuthorizationServer()
        server.init_app(app, query_client, save_token)
    else:
        server = AuthorizationServer(app, query_client, save_token)

    @app.route("/oauth/authorize", methods=["GET", "POST"])
    def authorize():
        user_id = request.values.get("user_id")
        if user_id:
            end_user = db.session.get(User, int(user_id))
        else:
            end_user = None

        try:
            grant = server.get_consent_grant(end_user=end_user)
        except OAuth2Error as error:
            return server.handle_error_response(request, error)

        if request.method == "GET":
            return grant.prompt or "ok"

        return server.create_authorization_response(grant=grant, grant_user=end_user)

    @app.route("/oauth/token", methods=["GET", "POST"])
    def issue_token():
        return server.create_token_response()

    return server


def create_flask_app():
    app = Flask(__name__)
    app.debug = True
    app.testing = True
    app.secret_key = "testing"
    app.config.update(
        {
            "SQLALCHEMY_TRACK_MODIFICATIONS": False,
            "SQLALCHEMY_DATABASE_URI": "sqlite://",
            "OAUTH2_ERROR_URIS": [
                ("invalid_client", "https://client.test/error#invalid_client")
            ],
        }
    )
    return app


def create_basic_header(username, password):
    text = f"{username}:{password}"
    auth = to_unicode(base64.b64encode(to_bytes(text)))
    return {"Authorization": "Basic " + auth}


def create_bearer_header(token):
    return {"Authorization": "Bearer " + token}



"""authlib.oauth2.rfc6749.authenticate_client.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Registry of client authentication methods, with 3 built-in methods:

1. client_secret_basic
2. client_secret_post
3. none

The "client_secret_basic" method is used a lot in examples of `RFC6749`_,
but the concept of naming are introduced in `RFC7591`_.

.. _`RFC6749`: https://tools.ietf.org/html/rfc6749
.. _`RFC7591`: https://tools.ietf.org/html/rfc7591
"""

import logging

from .errors import InvalidClientError
from .util import extract_basic_authorization

log = logging.getLogger(__name__)

__all__ = ["ClientAuthentication"]


class ClientAuthentication:
    def __init__(self, query_client):
        self.query_client = query_client
        self._methods = {
            "none": authenticate_none,
            "client_secret_basic": authenticate_client_secret_basic,
            "client_secret_post": authenticate_client_secret_post,
        }

    def register(self, method, func):
        self._methods[method] = func

    def authenticate(self, request, methods, endpoint):
        for method in methods:
            func = self._methods[method]
            client = func(self.query_client, request)
            if client and client.check_endpoint_auth_method(method, endpoint):
                request.auth_method = method
                return client

        if "client_secret_basic" in methods:
            raise InvalidClientError(
                status_code=401,
                description=f"The client cannot authenticate with methods: {methods}",
            )
        raise InvalidClientError(
            description=f"The client cannot authenticate with methods: {methods}",
        )

    def __call__(self, request, methods, endpoint="token"):
        return self.authenticate(request, methods, endpoint)


def authenticate_client_secret_basic(query_client, request):
    """Authenticate client by ``client_secret_basic`` method. The client
    uses HTTP Basic for authentication.
    """
    client_id, client_secret = extract_basic_authorization(request.headers)
    if client_id and client_secret:
        client = _validate_client(query_client, client_id, 401)
        if client.check_client_secret(client_secret):
            log.debug(f'Authenticate {client_id} via "client_secret_basic" success')
            return client
    log.debug(f'Authenticate {client_id} via "client_secret_basic" failed')


def authenticate_client_secret_post(query_client, request):
    """Authenticate client by ``client_secret_post`` method. The client
    uses POST parameters for authentication.
    """
    data = request.form
    client_id = data.get("client_id")
    client_secret = data.get("client_secret")
    if client_id and client_secret:
        client = _validate_client(query_client, client_id)
        if client.check_client_secret(client_secret):
            log.debug(f'Authenticate {client_id} via "client_secret_post" success')
            return client
    log.debug(f'Authenticate {client_id} via "client_secret_post" failed')


def authenticate_none(query_client, request):
    """Authenticate public client by ``none`` method. The client
    does not have a client secret.
    """
    client_id = request.payload.client_id
    if client_id and not request.payload.data.get("client_secret"):
        client = _validate_client(query_client, client_id)
        log.debug(f'Authenticate {client_id} via "none" success')
        return client
    log.debug(f'Authenticate {client_id} via "none" failed')


def _validate_client(query_client, client_id, status_code=400):
    if client_id is None:
        raise InvalidClientError(
            status_code=status_code,
            description="Missing 'client_id' parameter.",
        )

    client = query_client(client_id)
    if not client:
        raise InvalidClientError(
            status_code=status_code,
            description="The client does not exist on this server.",
        )

    return client



import pytest
from flask import current_app

from authlib.common.urls import add_params_to_uri
from authlib.common.urls import url_decode
from authlib.common.urls import urlparse
from authlib.jose import JsonWebToken
from authlib.oauth2.rfc6749.requests import BasicOAuth2Payload
from authlib.oauth2.rfc6749.requests import OAuth2Request
from authlib.oidc.core import ImplicitIDToken
from authlib.oidc.core.grants import OpenIDImplicitGrant as _OpenIDImplicitGrant

from .models import Client
from .models import exists_nonce

authorize_url = "/oauth/authorize?response_type=token&client_id=client-id"


@pytest.fixture(autouse=True)
def server(server):
    class OpenIDImplicitGrant(_OpenIDImplicitGrant):
        def get_jwt_config(self, client):
            alg = current_app.config.get("OAUTH2_JWT_ALG", "HS256")
            return dict(key="secret", alg=alg, iss="Authlib", exp=3600)

        def generate_user_info(self, user, scopes):
            return user.generate_user_info(scopes)

        def exists_nonce(self, nonce, request):
            return exists_nonce(nonce, request)

    server.register_grant(OpenIDImplicitGrant)
    return server


@pytest.fixture(autouse=True)
def client(client, db):
    client.set_client_metadata(
        {
            "redirect_uris": ["https://client.test/callback"],
            "scope": "openid profile",
            "token_endpoint_auth_method": "none",
            "response_types": ["id_token", "id_token token"],
        }
    )
    db.session.add(client)
    db.session.commit()
    return client


def validate_claims(id_token, params, alg="HS256"):
    jwt = JsonWebToken([alg])
    claims = jwt.decode(
        id_token, "secret", claims_cls=ImplicitIDToken, claims_params=params
    )
    claims.validate()
    return claims


def test_consent_view(test_client):
    rv = test_client.get(
        add_params_to_uri(
            "/oauth/authorize",
            {
                "response_type": "id_token",
                "client_id": "client-id",
                "scope": "openid profile",
                "state": "foo",
                "redirect_uri": "https://client.test/callback",
                "user_id": "1",
            },
        )
    )
    assert "error=invalid_request" in rv.location
    assert "nonce" in rv.location


def test_require_nonce(test_client):
    rv = test_client.post(
        "/oauth/authorize",
        data={
            "response_type": "id_token",
            "client_id": "client-id",
            "scope": "openid profile",
            "state": "bar",
            "redirect_uri": "https://client.test/callback",
            "user_id": "1",
        },
    )
    assert "error=invalid_request" in rv.location
    assert "nonce" in rv.location


def test_missing_openid_in_scope(test_client):
    rv = test_client.post(
        "/oauth/authorize",
        data={
            "response_type": "id_token token",
            "client_id": "client-id",
            "scope": "profile",
            "state": "bar",
            "nonce": "abc",
            "redirect_uri": "https://client.test/callback",
            "user_id": "1",
        },
    )
    assert "error=invalid_scope" in rv.location


def test_denied(test_client):
    rv = test_client.post(
        "/oauth/authorize",
        data={
            "response_type": "id_token",
            "client_id": "client-id",
            "scope": "openid profile",
            "state": "bar",
            "nonce": "abc",
            "redirect_uri": "https://client.test/callback",
        },
    )
    assert "error=access_denied" in rv.location


def test_authorize_access_token(test_client):
    rv = test_client.post(
        "/oauth/authorize",
        data={
            "response_type": "id_token token",
            "client_id": "client-id",
            "scope": "openid profile",
            "state": "bar",
            "nonce": "abc",
            "redirect_uri": "https://client.test/callback",
            "user_id": "1",
        },
    )
    assert "access_token=" in rv.location
    assert "id_token=" in rv.location
    assert "state=bar" in rv.location
    params = dict(url_decode(urlparse.urlparse(rv.location).fragment))
    validate_claims(params["id_token"], params)


def test_authorize_id_token(test_client):
    rv = test_client.post(
        "/oauth/authorize",
        data={
            "response_type": "id_token",
            "client_id": "client-id",
            "scope": "openid profile",
            "state": "bar",
            "nonce": "abc",
            "redirect_uri": "https://client.test/callback",
            "user_id": "1",
        },
    )
    assert "id_token=" in rv.location
    assert "state=bar" in rv.location
    params = dict(url_decode(urlparse.urlparse(rv.location).fragment))
    validate_claims(params["id_token"], params)


def test_response_mode_query(test_client):
    rv = test_client.post(
        "/oauth/authorize",
        data={
            "response_type": "id_token",
            "response_mode": "query",
            "client_id": "client-id",
            "scope": "openid profile",
            "state": "bar",
            "nonce": "abc",
            "redirect_uri": "https://client.test/callback",
            "user_id": "1",
        },
    )
    assert "id_token=" in rv.location
    assert "state=bar" in rv.location
    params = dict(url_decode(urlparse.urlparse(rv.location).query))
    validate_claims(params["id_token"], params)


def test_response_mode_form_post(test_client):
    rv = test_client.post(
        "/oauth/authorize",
        data={
            "response_type": "id_token",
            "response_mode": "form_post",
            "client_id": "client-id",
            "scope": "openid profile",
            "state": "bar",
            "nonce": "abc",
            "redirect_uri": "https://client.test/callback",
            "user_id": "1",
        },
    )
    assert b'name="id_token"' in rv.data
    assert b'name="state"' in rv.data


def test_client_metadata_custom_alg(test_client, app, db, client):
    """If the client metadata 'id_token_signed_response_alg' is defined,
    it should be used to sign id_tokens."""
    client.set_client_metadata(
        {
            "redirect_uris": ["https://client.test/callback"],
            "scope": "openid profile",
            "token_endpoint_auth_method": "none",
            "response_types": ["id_token", "id_token token"],
            "id_token_signed_response_alg": "HS384",
        }
    )
    db.session.add(client)
    db.session.commit()

    app.config["OAUTH2_JWT_ALG"] = None
    rv = test_client.post(
        "/oauth/authorize",
        data={
            "response_type": "id_token",
            "client_id": "client-id",
            "scope": "openid profile",
            "state": "foo",
            "redirect_uri": "https://client.test/callback",
            "user_id": "1",
            "nonce": "abc",
        },
    )
    params = dict(url_decode(urlparse.urlparse(rv.location).fragment))
    claims = validate_claims(params["id_token"], params, "HS384")
    assert claims.header["alg"] == "HS384"


def test_client_metadata_alg_none(test_client, app, db, client):
    """The 'none' 'id_token_signed_response_alg' alg should be
    forbidden in non implicit flows."""
    client.set_client_metadata(
        {
            "redirect_uris": ["https://client.test/callback"],
            "scope": "openid profile",
            "token_endpoint_auth_method": "none",
            "response_types": ["id_token", "id_token token"],
            "id_token_signed_response_alg": "none",
        }
    )
    db.session.add(client)
    db.session.commit()

    app.config["OAUTH2_JWT_ALG"] = None
    rv = test_client.post(
        "/oauth/authorize",
        data={
            "response_type": "id_token",
            "client_id": "client-id",
            "scope": "openid profile",
            "state": "foo",
            "redirect_uri": "https://client.test/callback",
            "user_id": "1",
            "nonce": "abc",
        },
    )
    params = dict(url_decode(urlparse.urlparse(rv.location).fragment))
    assert params["error"] == "invalid_request"


def test_deprecated_get_jwt_config_signature(user):
    """Using the old get_jwt_config(self) signature should emit a DeprecationWarning."""

    class DeprecatedImplicitGrant(_OpenIDImplicitGrant):
        def get_jwt_config(self):
            return {"key": "secret", "alg": "HS256", "iss": "Authlib", "exp": 3600}

        def generate_user_info(self, user, scopes):
            return user.generate_user_info(scopes)

        def exists_nonce(self, nonce, request):
            return exists_nonce(nonce, request)

    client = Client(
        user_id=user.id,
        client_id="deprecated-client",
        client_secret="secret",
    )
    client.set_client_metadata(
        {
            "redirect_uris": ["https://client.test/callback"],
            "scope": "openid profile",
            "token_endpoint_auth_method": "none",
            "response_types": ["id_token"],
        }
    )

    request = OAuth2Request("POST", "https://server.test/authorize")
    request.payload = BasicOAuth2Payload({"nonce": "test-nonce"})
    request.client = client
    request.user = user

    grant = DeprecatedImplicitGrant(request, client)
    token = {"scope": "openid", "expires_in": 3600}

    with pytest.warns(DeprecationWarning, match="get_jwt_config.*version 1.8"):
        grant.process_implicit_token(token)



"""
authlib.oauth2.rfc6749.grants
~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Implementation for `Section 4`_ of "Obtaining Authorization".

To request an access token, the client obtains authorization from the
resource owner. The authorization is expressed in the form of an
authorization grant, which the client uses to request the access
token. OAuth defines four grant types:

1. authorization code
2. implicit
3. resource owner password credentials
4. client credentials.

It also provides an extension mechanism for defining additional grant
types. Authlib defines refresh_token as a grant type too.

.. _`Section 4`: https://tools.ietf.org/html/rfc6749#section-4
"""

from .authorization_code import AuthorizationCodeGrant
from .base import AuthorizationEndpointMixin
from .base import BaseGrant
from .base import TokenEndpointMixin
from .client_credentials import ClientCredentialsGrant
from .implicit import ImplicitGrant
from .refresh_token import RefreshTokenGrant
from .resource_owner_password_credentials import ResourceOwnerPasswordCredentialsGrant

__all__ = [
    "BaseGrant",
    "AuthorizationEndpointMixin",
    "TokenEndpointMixin",
    "AuthorizationCodeGrant",
    "ImplicitGrant",
    "ResourceOwnerPasswordCredentialsGrant",
    "ClientCredentialsGrant",
    "RefreshTokenGrant",
]
